from typing import Set

from ..constants import PROBLEM_FEATURE, SAMPLE_FEATURE
from . import flacco, vm_features

FEATURES_FLACCO: Set = {
    SAMPLE_FEATURE.ELA_META,
    SAMPLE_FEATURE.ELA_DISTR,
    SAMPLE_FEATURE.ELA_LEVEL,
    SAMPLE_FEATURE.CM_ANGLE,
    SAMPLE_FEATURE.CM_CONV,
    SAMPLE_FEATURE.CM_GRAD,
    SAMPLE_FEATURE.CM_LIMO,
    SAMPLE_FEATURE.DISP,
    SAMPLE_FEATURE.IC,
    SAMPLE_FEATURE.NBC,
    SAMPLE_FEATURE.PCA,
    SAMPLE_FEATURE.FDC,
    PROBLEM_FEATURE.ELA_LOCAL,
    PROBLEM_FEATURE.ELA_CURV,
    PROBLEM_FEATURE.ELA_CONV,
    PROBLEM_FEATURE.LON,
    PROBLEM_FEATURE.HILL,
    PROBLEM_FEATURE.GRAD,
    PROBLEM_FEATURE.SCALE,
    PROBLEM_FEATURE.SOBOL,
}

FEATURES_VM: Set = {
    SAMPLE_FEATURE.ELA_META,
    SAMPLE_FEATURE.ELA_DISTR,
    SAMPLE_FEATURE.ELA_LEVEL,
}


def analyze_sample(feature_set: SAMPLE_FEATURE, *args, **kwargs):
    if feature_set in FEATURES_FLACCO:
        return getattr(flacco, feature_set)(*args, **kwargs)
    elif feature_set == SAMPLE_FEATURE.VARIABILITY_MAP:
        return vm_features.calculate(*args, **kwargs)
    else:
        raise ValueError(f"Unknown feature set {feature_set}")


def analyze_problem(feature_set: PROBLEM_FEATURE, *args, **kwargs):
    if feature_set in FEATURES_FLACCO:
        return getattr(flacco, feature_set)(*args, **kwargs)
    elif feature_set == SAMPLE_FEATURE.VARIABILITY_MAP:
        return vm_features.calculate(*args, **kwargs)
    else:
        raise ValueError(f"Unknown feature set {feature_set}")



from ..functions import cube
from ..problem import ProblemBase


class Problem(ProblemBase):
    def init(self):
        self.add_variable(bounds=(-10, 10), initial_guess=8)
        self.add_variable(bounds=(-10, 10), initial_guess=8)
        self.add_objective()
        self.add_solution(x=[1.0] * 2, f=0)

    def calc_objectives(self, x):
        return [cube(*x)]



from ..functions import himmelblau
from ..problem import ProblemBase


class Problem(ProblemBase):
    def init(self):
        self.add_variable(bounds=(-6, 6), initial_guess=6)
        self.add_variable(bounds=(-6, 6), initial_guess=6)
        self.add_objective()
        self.add_solution(x=[3, 2], f=0)

    def calc_objectives(self, x):
        return [himmelblau(*x)]



import numpy as np


def _type(func_type):
    """
    To classify functions:
      n - unknown,
      u - unimodal,
      m - multimodal,
      x - extreme multimodal
    """

    def wrapper(func):
        func.func_type = func_type
        return func

    return wrapper


# UNCONSTRAINED ############################################################


@_type("u")
def fon1(x):
    """
    Fonseca-Fleming's Problem (https://al-roomi.org/benchmarks/multi-objective/general-list/321-fonseca-fleming-s-function-fon)

    -4 <= xi <= 4

    """
    n = 1.0 / np.sqrt(len(x))
    f1 = 1 - np.exp(-np.sum([(xi - n) ** 2 for xi in x]))
    return f1


@_type("u")
def fon2(x):
    """
    Fonseca-Fleming's Problem (https://al-roomi.org/benchmarks/multi-objective/general-list/321-fonseca-fleming-s-function-fon)

    -4 <= xi <= 4

    """
    n = 1.0 / np.sqrt(len(x))
    f2 = 1 - np.exp(-np.sum([(xi + n) ** 2 for xi in x]))
    return f2


@_type("u")
def branin(x1, x2):
    """
    Branin Function (http://infinity77.net/global_optimization/test_functions_nd_B.html#go_benchmark.Branin01)

    -5 <= x1 <= 10
    0 <= x2 <= 15

    f(-pi, 12.275) = f(pi, 2.275) = f(9.4247, 2.475) = 0.3978
    """
    return (
        (x2 - (1.275 / np.pi**2) * x1**2 + (5.0 / np.pi) * x1 - 6) ** 2
        + (10 - (5.0 / (4.0 * np.pi))) * np.cos(x1)
        + 10
    )


@_type("m")
def easom(x):
    """
    Easom Function (http://infinity77.net/global_optimization/test_functions_nd_E.html#go_benchmark.Easom)

    -100 <= xi <= 100

    f(0) = 0
    """
    # It seems to be a wrong formula, see original at https://www.sfu.ca/~ssurjano/easom.html
    a = 20.0
    b = 0.2
    c = 2.0 * np.pi
    x = np.array(x)
    sum1 = a / np.exp(b * np.sqrt(np.mean(np.square(x))))
    sum2 = np.exp(np.mean(np.cos(c * x)))
    return a - sum1 + np.e - sum2


@_type("u")
def ellipsoid(x):
    """
    Rotated Ellipsoid Function

    -6 <= xi <= 6

    f(0) = 0
    """
    x = np.array([x], dtype=float, copy=False)
    # return (np.add.accumulate(x, axis=1)**2).sum(axis=1).reshape((-1, 1))
    return (np.add.accumulate(x, axis=1) ** 2).sum(axis=1)[0]


@_type("u")
def zerosum(x):
    """
    ZeroSum Function (http://infinity77.net/global_optimization/test_functions_nd_Z.html#go_benchmark.ZeroSum)

    -10 <= xi <= 10

    f(sum(x) == 0) = 0
    """
    if np.abs(np.sum(x)) < 3e-16:
        return 0.0
    return 1.0 + (10000.0 * np.abs(np.sum(x))) ** 0.5


@_type("m")
def wolfe(x1, x2, x3):
    """
    Gulf Function (http://infinity77.net/global_optimization/test_functions_nd_W.html#go_benchmark.Wolfe)

    0 <= xi <= 2

    f(0, 0, 0) = 0
    """
    return 4.0 / 3.0 * (x1**2 + x2**2 - x1 * x2) ** 0.75 + x3


@_type("x")
def whitley(x):
    """
    Whitley Function (http://infinity77.net/global_optimization/test_functions_nd_W.html#go_benchmark.Whitley)

    -10.24 <= xi <= 10.24

    f(1, ..., 1) = 0
    """
    xi = np.atleast_1d(x)
    xj = np.atleast_2d(x).T
    temp = 100.0 * ((xi**2.0) - xj) + (1.0 - xj) ** 2.0
    inner = (temp**2.0 / 4000.0) - np.cos(temp) + 1.0
    return np.sum(np.sum(inner, axis=0))


@_type("x")
def weierstrass(x):
    """
    Weierstrass Function (http://infinity77.net/global_optimization/test_functions_nd_W.html#go_benchmark.Weierstrass)

    -0.5 <= xi <= 0.5

    f(0, ..., 0) = 0
    """
    kmax = 20
    a, b = 0.5, 3.0
    x = np.atleast_1d(x)
    k = np.atleast_2d(np.arange(kmax + 1.0)).T
    t1 = a**k * np.cos(2 * np.pi * b**k * (x + 0.5))
    t2 = (len(x)) * np.sum(a**k.T * np.cos(np.pi * b**k.T))
    return np.sum(np.sum(t1, axis=0)) - t2


@_type("u")
def katsuura(x):
    """
    Katsuura Function (http://infinity77.net/global_optimization/test_functions_nd_K.html#go_benchmark.Katsuura)

    0 <= xi <= 100

    f(0.0, ..., 0.0) = 1.0
    """
    d = 32
    x = np.atleast_1d(x)
    k = np.atleast_2d(np.arange(1, d + 1)).T
    i = np.arange(0.0, (len(x)) * 1.0)
    inner = np.floor(2**k * x) * (2.0 ** (-k))
    return np.prod(np.sum(inner, axis=0) * (i + 1) + 1)


@_type("x")
def keane(x1, x2):
    """
    Keane Function (http://infinity77.net/global_optimization/test_functions_nd_K.html#go_benchmark.Keane)

    0 <= xi <= 10

    f(0.0, 1.39325) = -0.6737
    """
    delimiter = np.sqrt(x1**2 + x2**2)
    if delimiter == 0:
        return np.nan
    return -((np.sin(x1 - x2)) ** 2 * (np.sin(x1 + x2)) ** 2) / delimiter


@_type("m")
def hosaki(x1, x2):
    """
    Hosaki Function (http://infinity77.net/global_optimization/test_functions_nd_H.html#go_benchmark.Hosaki)

    0 <= xi <= 10

    f(4, 2) = -2.34589
    """
    return (
        (1 - 8 * x1 + 7 * x1**2 - (7.0 / 3.0) * x1**3 + 0.25 * x1**4)
        * x2**2
        * np.exp(-x2)
    )


@_type("m")
def himmelblau(x1, x2):
    """
    HimmelBlau Function (http://infinity77.net/global_optimization/test_functions_nd_H.html#go_benchmark.HimmelBlau)

    -6 <= xi <= 6

    f(0, 0) = 0
    """
    return (x1**2 + x2 - 11) ** 2 + (x1 + x2**2 - 7) ** 2


@_type("m")
def gulf(x1, x2, x3):
    """
    Gulf Function (http://infinity77.net/global_optimization/test_functions_nd_S.html#go_benchmark.Sargan)

    0.1 <= xi <= 100

    f(50, 25, 1.5) = 0
    """
    m = 99.0
    i = np.arange(1.0, m + 1)
    u = 25 + (-50 * np.log(i / 100.0)) ** (2 / 3.0)
    vec = np.exp(-((np.abs(u - x2)) ** x3 / x1)) - i / 100.0
    return sum(vec**2)


@_type("u")
def sargan(x):
    """
    Sargan Function (http://infinity77.net/global_optimization/test_functions_nd_S.html#go_benchmark.Sargan)

    -100 <= xi <= 100

    f(0, ..., 0) = 0
    """
    result = 0
    for i in np.arange(len(x)):
        result += len(x) * (
            x[i] ** 2 + 0.4 * np.sum([x[i] * x[j] for j in np.arange(len(x)) if i != j])
        )
    return result


@_type("u")
def sodp(x):
    """
    Sum Of Different Powers Function (http://infinity77.net/global_optimization/test_functions_nd_S.html#go_benchmark.Sodp)

    -1 <= xi <= 1

    f(0, ..., 0) = 0
    """
    return np.sum([(np.abs(x[i])) ** (i + 2) for i in np.arange(len(x))])


@_type("u")
def step(x):
    """
    Step Function (http://infinity77.net/global_optimization/test_functions_nd_S.html#go_benchmark.Step)

    -100 <= xi <= 100

    f(0, ..., 0) = 0
    """
    return np.sum([(np.floor(xi)) ** 2 for xi in x])


@_type("m")
def multimodal(x):
    """
    MultiModal Function (http://infinity77.net/global_optimization/test_functions_nd_M.html#go_benchmark.MultiModal)

    -10 <= xi <= 10

    f(0, ..., 0) = 0
    """
    return np.sum(np.abs(x)) * np.prod(np.abs(x))


@_type("m")
def mishra7(x):
    """
    Exponential Function (http://infinity77.net/global_optimization/test_functions_nd_M.html#go_benchmark.Mishra07)

    -10 <= xi <= 10

    f(sqrt(n), ..., sqrt(n)) = 0
    """
    return (np.prod(x) - np.prod(np.arange(1, len(x) + 1))) ** 2.0


@_type("u")
def exponential(x):
    """
    Exponential Function (http://infinity77.net/global_optimization/test_functions_nd_E.html#go_benchmark.Exponential)

    -1 <= xi <= 1

    f(0, ..., 0) = -1
    """
    return -np.exp(-0.5 * np.sum([xi**2 for xi in x]))


@_type("u")
def exp2(x1, x2):
    """
    Exp2 Function (http://infinity77.net/global_optimization/test_functions_nd_E.html#go_benchmark.Exp2)

    0 <= xi <= 20

    f(1.0, 10.0) = 0
    """

    def sum_e(i):
        t = -0.1 * i
        return np.exp(t * x1) - 5 * np.exp(t * x2) - np.exp(t) + 5 * np.exp(-i)

    return np.sum([sum_e(i) ** 2 for i in range(10)])


@_type("u")
def elatar(x1, x2):
    """
    El-Attar-Vidyasagar-Dutta Function (http://infinity77.net/global_optimization/test_functions_nd_D.html#go_benchmark.DeckkersAarts)

    -100 <= xi <= 100

    f(3.409186, -2.171433) = 1.712780354
    """
    return (
        (x1**2 + x2 - 10) ** 2
        + (x1 + x2**2 - 7) ** 2
        + (x1**2 + x2**3 - 1) ** 2
    )


@_type("u")
def dixonprice(x):
    """
    DixonPrice Function (http://infinity77.net/global_optimization/test_functions_nd_D.html#go_benchmark.DixonPrice)

    -10 <= xi <= 10

    f(x1, ..., xn) = 0, xi = 2.0 ** ((2.0 - 2.0 ** i) / 2.0 ** i)
    """
    return (x[0] - 1.0) ** 2 + np.sum(
        [(i + 1.0) * (2 * x[i] ** 2 - x[i - 1]) ** 2 for i in range(1, len(x))]
    )


@_type("u")
def deckkersaarts(x1, x2):
    """
    Deckkers-Aarts Function (http://infinity77.net/global_optimization/test_functions_nd_D.html#go_benchmark.DeckkersAarts)

    -20 <= xi <= 20

    f(0, +-15) = -24777
    """
    return (
        1e5 * x1**2
        + x2**2
        - (x1**2 + x2**2) ** 2
        + 1e-5 * (x1**2 + x2**2) ** 4
    )


@_type("m")
def deceptive(x, a=0.3):
    """
    Deceptive Function (http://infinity77.net/global_optimization/test_functions_nd_D.html#go_benchmark.Deceptive)

    0 <= xi <= 1

    f(a1, ..., aN) = -1
    """
    if (np.array(x) < 0).any() | (1 < np.array(x)).any():
        return np.nan
    a = [a] * len(x)

    def g(i):
        if 0 <= x[i] <= (4.0 / 5.0) * a[i]:
            return -x[i] / a[i] + (4.0 / 5.0)
        elif (4.0 / 5.0) * a[i] <= x[i] <= a[i]:
            return 5.0 * x[i] / a[i] - 4.0
        elif a[i] <= x[i] <= (1.0 + 4.0 * a[i]) / 5.0:
            return 5.0 * (x[i] - a[i]) / (a[i] - 1)
        elif (1.0 + 4.0 * a[i]) / 5.0 <= x[i] <= 1:
            return (x[i] - 1) / (1 - a[i])

    return -np.mean([g(i) for i in range(len(x))]) ** 2


@_type("u")
def decanomial(x1, x2):
    """
    Decanomial Function (http://infinity77.net/global_optimization/test_functions_nd_D.html#go_benchmark.Decanomial)

    -10 <= xi <= 10

    f(2, -3) = 0
    """
    s1 = abs(x2**4 + 12.0 * x2**3 + 54.0 * x2**2 + 108.0 * x2 + 81.0)
    s2 = abs(
        x1**10
        - 20.0 * x1**9
        + 180.0 * x1**8
        - 960.0 * x1**7
        + 3360.0 * x1**6
        - 8064.0 * x1**5
        + 13340.0 * x1**4
        - 15360.0 * x1**3
        + 11520.0 * x1**2
        - 5120.0 * x1
        + 2624.0
    )
    return 0.001 * (s1 + s2) ** 2


@_type("u")
def cube(x1, x2):
    """
    Cube Function (http://infinity77.net/global_optimization/test_functions_nd_C.html#go_benchmark.Cube)

    -10 <= xi <= 10

    f(1, 1) = 0
    """
    return 100 * (x2 - x1**3) ** 2 + (1 - x1) ** 2


@_type("u")
def cigar(x):
    """
    Cigar Function (http://infinity77.net/global_optimization/test_functions_nd_C.html#go_benchmark.Cigar)

    -100 <= xi <= 100

    f(0, ..., 0) = 0
    """
    return x[0] ** 2 + 1e6 * np.sum([x[i] ** 2 for i in range(1, len(x))])


@_type("u")
def bukin06(x1, x2):
    """
    Bukin06 Function (http://infinity77.net/global_optimization/test_functions_nd_B.html#go_benchmark.Bukin06)

    -15 <= x1 <= -5
    -3 <= x2 <= 3

    f(-10, 1) = 0
    """
    return 100.0 * np.sqrt(abs(x2 - 0.01 * x1**2)) + 0.01 * abs(x1 + 10)


@_type("u")
def bukin04(x1, x2):
    """
    Bukin04 Function (http://infinity77.net/global_optimization/test_functions_nd_B.html#go_benchmark.Bukin04)

    -15 <= x1 <= -5
    -3 <= x2 <= 3

    f(-10, 0) = 0
    """
    return 100.0 * x2**2 + 0.01 * abs(x1 + 10.0)


@_type("u")
def bukin02(x1, x2):
    """
    Bukin02 Function (http://infinity77.net/global_optimization/test_functions_nd_B.html#go_benchmark.Bukin02)

    -15 <= x1 <= -5
    -3 <= x2 <= 3

    f(-10, 0) = 0
    """
    return 100.0 * (x2**2 - 0.01 * x1**2 + 1.0) + 0.01 * (x1 + 10.0) ** 2


@_type("u")
def brown(x):
    """
    Brown Function (http://infinity77.net/global_optimization/test_functions_nd_B.html#go_benchmark.Brown)

    -1 <= xi <= 4

    f(0, ..., 0) = 0
    """
    return np.sum(
        [
            (x[i] ** 2) ** (x[i + 1] ** 2 + 1) + (x[i + 1] ** 2) ** (x[i] ** 2 + 1)
            for i in range(len(x) - 1)
        ]
    )


@_type("u")
def brent(x1, x2):
    """
    Brent Function (http://infinity77.net/global_optimization/test_functions_nd_B.html#go_benchmark.Brent)

    -10 <= xi <= 10

    f(-10, -10) = 0
    """
    return (x1 + 10.0) ** 2 + (x2 + 10.0) ** 2 + np.exp(-(x1**2) - x2**2)


@_type("u")
def beale(x1, x2):
    """
    Beale Function (http://infinity77.net/global_optimization/test_functions_nd_B.html#go_benchmark.Beale)

    -10 <= xi <= 10

    f(3.0, 0.5) = 0
    """
    return (
        (x1 * x2 - x1 + 1.5) ** 2
        + (x1 * x2**2 - x1 + 2.25) ** 2
        + (x1 * x2**3 - x1 + 2.652) ** 2
    )


@_type("u")
def amgm(x):
    """
    AMGM Function (http://infinity77.net/global_optimization/test_functions_nd_A.html#go_benchmark.AMGM)
    Arithmetic Mean - Geometric Mean Equality

    0 <= xi <= 10

    f(0, ..., 0) = 0
    """
    power = 1.0 / len(x)
    if np.any(np.array(x) < 0):
        return np.nan
    return (power * np.sum(x) - np.prod(x) ** power) ** 2


@_type("x")
def ackley(x):
    """
    Ackley Function (http://benchmarkfcns.xyz/benchmarkfcns/ackleyfcn.html)

    -2 <= xi <= 2

    f(0, ..., 0) = 0
    """
    firstSum = 0.0
    secondSum = 0.0
    for xi in x:
        firstSum += xi**2
        secondSum += np.cos(2.0 * np.pi * xi)
    return (
        -20.0 * np.exp(-0.2 * np.sqrt(firstSum / len(x)))
        - np.exp(secondSum / len(x))
        + 20
        + np.e
    )


@_type("m")
def goldsteinprice(x1, x2):
    """
    Goldstein-Price's Function (https://www.sfu.ca/~ssurjano/goldpr.html)

    -2 <= x1 <= 2
    -2 <= x2 <= 2

    f(0, -1) = 3
    """
    f1 = (x1 + x2 + 1) ** 2
    f2 = 19 - 14 * x1 + 3 * x1**2 - 14 * x2 + 6 * x1 * x2 + 3 * x2**2
    f3 = (2 * x1 - 3 * x2) ** 2
    f4 = 18 - 32 * x1 + 12 * x1**2 + 48 * x2 - 36 * x1 * x2 + 27 * x2**2
    return (1 + f1 * f2) * (30 + f3 * f4)


@_type("u")
def rosenbrock(x):
    """
    Rosenbrock's Valley Function (https://en.wikipedia.org/wiki/Rosenbrock_function)

    -2.048 <= xi <= 2.048

    f(1, ..., 1) = 0
    """
    result = 0
    for i in range(len(x) - 1):
        result += 100 * (x[i + 1] - x[i] ** 2) ** 2 + (1 - x[i]) ** 2
    return result


@_type("m")
def sixhumpcamel(x1, x2):
    """
    Six Hump Camel Back Function (https://www.sfu.ca/~ssurjano/camel6.html)

    -3 <= x1 <= 3
    -2 <= x2 <= 2

    f(0.0898, -0.7126) = f(-0.0898, 0.7126) = -1.03164
    """
    f1 = (4 - 2.1 * x1**2 + x1**4 / 3.0) * x1**2
    f2 = x1 * x2
    f3 = (-4 + 4 * x2**2) * x2**2
    return f1 + f2 + f3


@_type("x")
def griewank(x):
    """
    Griewank Function (https://www.sfu.ca/~ssurjano/griewank.html)

    -6 <= xi <= 6

    f(0, ..., 0) = 0
    """
    return (
        1.0
        + np.sum([x[i] ** 2 for i in range(len(x))]) / 4000.0
        - np.prod([np.cos(x[i] / np.sqrt(i + 1)) for i in range(len(x))])
    )


@_type("x")
def levy3(x):
    """
    Levy 3 Function (http://infinity77.net/global_optimization/test_functions_nd_L.html#go_benchmark.Levy03)

    -10 <= xi <= 10

    f(1, ..., 1) = 0
    """
    x = [1 + (xi - 1) / 4 for xi in x]
    return (
        np.sin(np.pi * x[0]) ** 2
        + (x[-1] - 1) ** 2 * (1 + np.sin(2 * np.pi * x[-1]) ** 2)
        + np.sum(
            [
                (x[i] - 1) ** 2 * (1 + 10 * np.sin(np.pi * x[i] + 1) ** 2)
                for i in range(len(x) - 1)
            ]
        )
    )


@_type("x")
def levy5(x1, x2):
    """
    Levy 5 Function (http://infinity77.net/global_optimization/test_functions_nd_L.html#go_benchmark.Levy05)

    -2 <= xi <= 2

    f(-1.3086, -1.4248) = -176.1375
    """
    return (
        np.sum([i * np.cos(x1 * (i - 1) + i) for i in range(1, 6)])
        * np.sum([j * np.cos(x2 * (j + 1) + j) for j in range(1, 6)])
        + (x1 + 1.42513) ** 2
        + (x2 + 0.80032) ** 2
    )


@_type("x")
def rastrigin(x):
    """
    Rastrigin Function (http://www.sfu.ca/~ssurjano/rastr.html)

    -5.12 <= xi <= 5.12

    f(0, ..., 0) = 0
    """
    return 10 * len(x) + np.sum(
        [x[i] * x[i] - 10.0 * np.cos(2.0 * x[i] * np.pi) for i in range(len(x))]
    )


@_type("u")
def perm(x):
    """
    Very sensitive to x
    Perm Function (http://www.sfu.ca/~ssurjano/permdb.html,
                   http://infinity77.net/global_optimization/test_functions_nd_P.html#go_benchmark.PermFunction01)

    -d <= xi <= d

    f(1, 2, .., dim) = 0
    """
    b = 0.5
    result = 0
    for i in range(1, len(x) + 1):
        inner_sum = np.sum(
            [(j**i + b) * ((x[j - 1] / j) ** i - 1.0) for j in range(1, len(x) + 1)]
        )
        result += inner_sum**2
    return result


@_type("u")
def perm2(x):
    """
    Perm Function (http://www.sfu.ca/~ssurjano/perm0db.html)

    -d <= xi <= d

    f(1, 1/2, .., 1/dim) = 0
    """
    b = 10
    result = 0
    for i in range(1, len(x) + 1):
        inner_sum = np.sum(
            [(j + b) * (x[j - 1] ** i - 1.0 / j**i) for j in range(1, len(x) + 1)]
        )
        result += inner_sum**2
    return result


@_type("u")
def powellsum(x):
    """
    Sum of Different Powers Function (http://benchmarkfcns.xyz/benchmarkfcns/powellsumfcn.html)

    -5 <= xi <= 5

    f(0, ..., 0) = 0
    """
    return np.sum([np.abs(x[i]) ** (i + 2) for i in range(len(x))])


@_type("u")
def powell(xx, x1=None, x2=None, x3=None):  # dummy variables for parsing
    """
    Powell function (http://www.sfu.ca/~ssurjano/powell.html)

    -4 <= xi <= 5

    f(0, ..., 0) = 0
    """
    if len(xx) % 4 != 0:
        raise ValueError("The problem dimension should be multiple of 4")

    sum = 0
    for i in range(len(xx) // 4):
        i *= 4
        sum += (xx[i] + 10 * xx[i + 1]) ** 2
        sum += 5 * (xx[i + 2] - xx[i + 3]) ** 2
        sum += (xx[i + 1] - 2 * xx[i + 2]) ** 4
        sum += 10 * (xx[i] - xx[i + 3]) ** 4
    return sum


@_type("u")
def trid(x):
    """
    Trid Function (https://www.sfu.ca/~ssurjano/trid.html)

    -dim**2 <= xi <= dim**2

    dim = 6: f(6, 10, 12, 12, 10, 6) = -50
    """
    return np.sum([(x[i] - 1) ** 2 for i in range(len(x))]) - np.sum(
        [x[i] * x[i - 1] for i in range(1, len(x))]
    )


@_type("u")
def zakharov(x):
    """
    Zakharov Function (http://www.sfu.ca/~ssurjano/zakharov.html)

    -5 <= xi <= 10

    f(0, ..., 0) = 0
    """
    sum1 = np.sum([x[i] ** 2 for i in range(len(x))])
    sum2 = np.sum([0.5 * (i + 1) * x[i] for i in range(len(x))])
    return sum1 + sum2**2 + sum2**4


@_type("x")
def michalewicz(x):
    """
    Michalewicz Function (http://www.sfu.ca/~ssurjano/michal.html)

    0 <= xi <= pi

    dim = 2: f(2.20, 1.57) = -1.8013
    dim = 5: f = -4.687658
    dim = 10: f = -9.66015
    """

    return -np.sum(
        [
            np.sin(x[i]) * (np.sin((i + 1) * x[i] ** 2 / np.pi) ** 20)
            for i in range(len(x))
        ]
    )


@_type("m")
def styblinskitang(x):
    """
    Styblinski-Tang Function (http://www.sfu.ca/~ssurjano/stybtang.html)

    -5 <= xi <= 5

    f(-2.903534, ..., -2.903534) = -39.166 * d
    """
    return 0.5 * np.sum([x[i] ** 4 - 16 * x[i] ** 2 + 5 * x[i] for i in range(len(x))])


@_type("x")
def shubert4(x):
    """
    Shubert 4 Function (http://benchmarkfcns.xyz/benchmarkfcns/shubert4fcn.html)

    -2 <= xi <= 2

    f = -29.017
    """
    return np.sum(
        [
            np.sum([j * np.cos((j + 1) * x[i] + j) for j in range(1, 6)])
            for i in range(len(x))
        ]
    )


@_type("x")
def shubert(x):
    """
    Shubert Function (http://benchmarkfcns.xyz/benchmarkfcns/shubertfcn.html)

    -2 <= xi <= 2

    f = -186.7309
    """
    return np.prod(
        [
            np.sum([j * np.cos((j + 1) * x[i] + j) for j in range(1, 6)])
            for i in range(len(x))
        ]
    )


@_type("x")
def salomon(x):
    """
    Salomon Function (http://benchmarkfcns.xyz/benchmarkfcns/salomonfcn.html)

    -2 <= xi <= 2

    f(0, ..., 0) = 0
    """
    s = np.sqrt(np.sum([xi**2 for xi in x]))
    return 1 - np.cos(2 * np.pi * s) + 0.1 * s


@_type("x")
def bohachevsky(x):
    """
    Bohachevsky Function (http://infinity77.net/global_optimization/test_functions_nd_B.html#go_benchmark.Bohachevsky)

    -2 <= xi <= 2

    f(0, ..., 0) = 0
    """
    s = 0
    for i in range(len(x) - 1):
        s += (
            x[i] ** 2
            + 2 * x[i + 1] ** 2
            - 0.3 * np.cos(3 * np.pi * x[i])
            - 0.4 * np.cos(4 * np.pi * x[i + 1])
            + 0.7
        )
    return s


@_type("u")
def csendes(x):
    """
    Csendes Function (http://infinity77.net/global_optimization/test_functions_nd_C.html#go_benchmark.Csendes)

    -1 <= xi <= 1

    f(0, ..., 0) = 0
    """
    s = 0
    for i in range(len(x)):
        if x[i] == 0:
            return float("nan")
        s += x[i] ** 6 * (2 + np.sin(1.0 / x[i]))
    return s


@_type("m")
def quintic(x):
    """
    Quintic Function (http://infinity77.net/global_optimization/test_functions_nd_Q.html#go_benchmark.Quintic)

    -10 <= xi <= 10

    f(-1, ..., -1) = 0
    """
    s = 0
    for xi in x:
        s += np.abs(xi**5 - 3 * xi**4 + 4 * xi**3 + 2 * xi**2 - 10 * xi - 4)
    return s


@_type("x")
def xinsheyang4(x):
    """
    Xin-She Yang 4 Function (http://benchmarkfcns.xyz/benchmarkfcns/xinsheyangn4fcn.html)

    -6 <= xi <= 6

    f(0, ..., 0) = -1
    """
    sum1 = 0
    sum2 = 0
    sum3 = 0
    for xi in x:
        sum1 += np.sin(xi) ** 2
        sum2 += xi**2
        sum3 += np.sin(np.sqrt(np.abs(xi))) ** 2
    return (sum1 - np.exp(-sum2)) * np.exp(-sum3)


@_type("x")
def xinsheyang2(x):
    """
    Xin-She Yang 2 Function (http://benchmarkfcns.xyz/benchmarkfcns/xinsheyangn2fcn.html)

    -2pi <= xi <= 2pi

    f(0, ..., 0) = 0
    """
    sum1 = 0
    sum2 = 0
    for xi in x:
        sum1 += np.abs(xi)
        sum2 += np.sin(xi**2)
    return sum1 * np.exp(-sum2)


@_type("u")
def qing(x):
    """
    Qing Function (http://benchmarkfcns.xyz/benchmarkfcns/qingfcn.html)

    -500 <= xi <= 500

    f(+-sqrt(i), ..., +-sqrt(i)) = 0
    """
    return np.sum([(x[i] ** 2 - i - 1) ** 2 for i in range(len(x))])



import logging
from pathlib import Path
from typing import Callable, List, Tuple

import joblib
import numpy as np
import pandas as pd
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline

from .problems.problem import ProblemBase
from .sample import Sample

LOG = logging.getLogger(__name__)


def load_predictor(
    problem: ProblemBase,
    path: Path,
) -> Callable[[np.ndarray, np.ndarray], Tuple[float, float]]:
    predictor = joblib.load(path)

    def predict_nu(x: np.ndarray, f: np.ndarray):
        sample = Sample(id="", problem=problem, x=x, f=f)
        features_dict = sample.features()
        features = pd.DataFrame({f: [v] for f, v in features_dict.items()})
        features = features.replace([np.inf, -np.inf], np.nan)
        nu = float(predictor.predict(features)[0])
        return nu

    return predict_nu


def build_predictor(
    sample: pd.DataFrame, columns_x: List[str], column_f: str
) -> Tuple[Pipeline, dict]:
    vm_features = [_ for _ in columns_x if _.startswith("feature.vm.")]
    flacco_features = [_ for _ in columns_x if not _.startswith("feature.vm.")]
    len(columns_x), len(vm_features), len(flacco_features)

    predictor = Pipeline(
        [
            ("imputer", SimpleImputer(missing_values=np.nan, strategy="mean")),
            ("classifier", RandomForestClassifier(random_state=0, n_jobs=8)),
        ]
    )

    sample[columns_x] = sample[columns_x].replace([np.inf, -np.inf], np.nan)

    best_score = -np.inf
    best_depth = None
    for max_depth in np.arange(4, 24, 2):
        predictor.set_params(classifier__max_depth=max_depth)
        score_train, score_test = cross_validate(
            sample=sample,
            columns_x=columns_x,
            column_f=column_f,
            predictor=predictor,
        )
        LOG.debug(f"{max_depth=} {score_train=:.3f} {score_test=:.3f}")
        if score_test > best_score:
            best_score = score_test
            best_depth = max_depth

    sample_x = sample[columns_x]
    sample_f = sample[column_f]
    predictor.set_params(classifier__max_depth=best_depth)
    predictor.fit(sample_x, sample_f)
    score_train = predictor.score(sample_x, sample_f)
    LOG.info(f"  {best_depth=:<4} {best_score=:.3f} {score_train=:.3f}")
    scores = {
        "score_train": score_train,
        "score_test": np.nan,
        "score_cv": best_score,
        "depth_cv": int(best_depth),
    }
    return predictor, scores


def cross_validate(
    sample: pd.DataFrame,
    columns_x: List[str],
    column_f: str,
    predictor: Pipeline,
    cv_count: int = 5,
):
    problems_all = sample["problem"].drop_duplicates().values
    cv_scores_train = []
    cv_scores_test = []
    for problems_test in np.array_split(problems_all, cv_count, axis=0):
        masks = [sample["problem"] == p_name for p_name in problems_test]
        test_mask = np.logical_or.reduce(masks)
        train_x = sample.loc[~test_mask, columns_x]
        train_f = sample.loc[~test_mask, column_f]
        test_x = sample.loc[test_mask, columns_x]
        test_f = sample.loc[test_mask, column_f]
        predictor.fit(train_x, train_f)
        cv_scores_train.append(predictor.score(train_x, train_f))
        cv_scores_test.append(predictor.score(test_x, test_f))
    return np.mean(cv_scores_train), np.mean(cv_scores_test)



from ..functions import exp2
from ..problem import ProblemBase


class Problem(ProblemBase):
    def init(self):
        self.add_variable(bounds=(0, 20), initial_guess=18)
        self.add_variable(bounds=(0, 20), initial_guess=18)
        self.add_objective()
        self.add_solution(x=[1.0, 10], f=0.0)

    def calc_objectives(self, x):
        return [exp2(*x)]



from enum import StrEnum
from typing import Self


class METRICS(StrEnum):
    NONE = "none"
    PREDICTED = "predicted"

    R2_CV = "R2_cv"
    SUIT_EXT = "suitability_ext"
    VM_ANGLES_EXT = "vm_angles_ext"
    SEGMENT_EXT = "segment_ext"
    SUGGESTED_F = "suggested_f"


METRICS_MAX = {
    METRICS.R2_CV,
    METRICS.SUIT_EXT,
    METRICS.SUGGESTED_F,
}
METRICS_MIN = {
    METRICS.VM_ANGLES_EXT,
    METRICS.SEGMENT_EXT,
}


class SAMPLE_FEATURE(StrEnum):
    # Sample-based features
    ELA_META = "ela_meta"
    ELA_DISTR = "ela_distr"
    ELA_LEVEL = "ela_level"
    CM_ANGLE = "cm_angle"
    CM_CONV = "cm_conv"
    CM_GRAD = "cm_grad"
    CM_LIMO = "cm_limo"
    DISP = "disp"
    IC = "ic"
    NBC = "nbc"
    PCA = "pca"
    FDC = "fdc"
    VARIABILITY_MAP = "variability_map"

    @staticmethod
    def bounds_required(feature_set: Self):
        return feature_set in {
            SAMPLE_FEATURE.CM_ANGLE,
            SAMPLE_FEATURE.CM_CONV,
            SAMPLE_FEATURE.CM_GRAD,
            SAMPLE_FEATURE.CM_LIMO,
        }


class PROBLEM_FEATURE(StrEnum):
    # Features that require additional evaluations
    ELA_LOCAL = "ela_local"
    ELA_CURV = "ela_curv"
    ELA_CONV = "ela_conv"
    LON = "lon"
    HILL = "hill"
    GRAD = "grad"
    SCALE = "scale"
    SOBOL = "sobol"

    @staticmethod
    def sample_required(feature_set: Self):
        return feature_set in {
            PROBLEM_FEATURE.ELA_LOCAL,
            PROBLEM_FEATURE.ELA_CURV,
            PROBLEM_FEATURE.ELA_CONV,
        }

    @staticmethod
    def bounds_required(feature_set: Self):
        return feature_set in {
            PROBLEM_FEATURE.ELA_LOCAL,
            PROBLEM_FEATURE.ELA_CURV,
            PROBLEM_FEATURE.LON,
            PROBLEM_FEATURE.HILL,
            PROBLEM_FEATURE.GRAD,
            PROBLEM_FEATURE.SCALE,
            PROBLEM_FEATURE.SOBOL,
        }



import inspect
from pathlib import Path

_all = []
_classes = set([])
__all__ = [
    name.stem
    for name in Path(__file__).parent.iterdir()
    if name.suffix == ".py" and not name.stem.startswith("_")
]


def add_problem(class_name: str, func_name: str):
    if class_name in globals():
        return f"{class_name}.append({func_name})"
    else:
        _classes.add(class_name)
        return f"{class_name} = [{func_name}]"


for _name in __all__:
    exec(f"from . import {_name}")
    exec(f"_all.append({_name})")
    # Classify problem by multi-modality (_u, _m, _x)
    exec(f"from ..functions import {_name} as _function")
    exec(add_problem(class_name=f"_t{_function.func_type}", func_name=_name))
    # Classify problem by dimension (_d2, _d3, ..., _dn)
    args = [arg for arg in inspect.getfullargspec(_function).args if "x" in arg]
    if ["x"] == args:
        exec(add_problem(class_name="_dn", func_name=_name))
    else:
        exec(add_problem(class_name=f"_d{len(args)}", func_name=_name))

# Sort by name
_all = sorted(_all, key=lambda problem: problem.Problem().NAME)
for class_name in _classes:
    exec(f"{class_name}.sort(key=lambda problem: problem.Problem().NAME)")

# Check that all problems were classified
assert len(__all__) == sum(eval(f"len({c})") for c in _classes if c.startswith("_d"))
assert len(__all__) == sum(eval(f"len({c})") for c in _classes if c.startswith("_t"))



import json
import logging
from dataclasses import asdict, dataclass
from enum import StrEnum
from pathlib import Path
from typing import Dict, List, Optional, Tuple

import numpy as np

from .constants import METRICS, SAMPLE_FEATURE
from .problems.problem import ProblemBase

LOG = logging.getLogger(__name__)


class DOE(StrEnum):
    FF = "sample_ff"
    LHS = "sample_lhs"
    GRID = "sample_grid"
    RAND = "sample_rand"


class MODE(StrEnum):
    AUTO = "auto"
    LOAD = "load"
    GENERATE = "generate"


@dataclass
class ModelInfo:
    nu: float
    length_scale: float
    metric_value: float


def clear_samples_cache():
    path = Path(__file__).parent / "samples"
    for p_path in path.iterdir():
        LOG.debug(f"Deleting {p_path}")
        for s_path in p_path.iterdir():
            s_path.unlink()
        p_path.rmdir()


@dataclass
class Sample:
    id: str
    problem: ProblemBase
    x: np.ndarray
    f: np.ndarray

    def features(self) -> Dict[str, float]:
        features = {
            **self._features(SAMPLE_FEATURE.ELA_META),
            **self._features(SAMPLE_FEATURE.ELA_DISTR),
            **self._features(SAMPLE_FEATURE.DISP),
            **self._features(SAMPLE_FEATURE.IC),
            **self._features(SAMPLE_FEATURE.PCA),
            **self._features(SAMPLE_FEATURE.VARIABILITY_MAP),
        }
        # Remove *.const_runtime features to make results reproducible:
        # - 'disp.costs_runtime'
        # - 'ic.costs_runtime'
        # - 'pca.costs_runtime'
        return {
            f"feature.{f}": val
            for f, val in features.items()
            if "costs_runtime" not in f
        }

    def _features(self, feature: SAMPLE_FEATURE, *args, **kwargs) -> Dict[str, float]:
        return _features_estimate(
            id=self.id,
            x=self.x,
            f=self.f,
            bounds=self.problem.variable_bounds,
            feature=feature,
            *args,
            **kwargs,
        )

    def models_by_metric(self) -> Dict[METRICS, List[ModelInfo]]:
        return {
            metric: _models_estimate(
                id=self.id,
                problem=self.problem,
                x=self.x,
                f=self.f,
                metric=metric,
            )
            for metric in [
                METRICS.R2_CV,
                METRICS.SUIT_EXT,
                METRICS.VM_ANGLES_EXT,
                METRICS.SEGMENT_EXT,
            ]
        }


class SampleCached:
    def __init__(
        self,
        problem: ProblemBase,
        size: int,
        doe: DOE = DOE.RAND,
        mode: MODE = MODE.AUTO,
        options: Optional[dict] = None,
        tag: str = "",
    ):
        options = options or {}
        options["size"] = size
        doe_args_str = " ".join([f"{k}={options[k]}" for k in sorted(options)])

        problem_id = f"{problem.NAME}_{problem.dim_x}_{problem.dim_f}"
        doe_id = f"{doe.lower()} {doe_args_str}"
        self.id = f"{problem_id} {doe_id}"
        self.problem = problem

        file_path = Path(__file__).parent / "samples" / problem_id
        if tag:
            file_path = file_path / tag
        self.path = file_path / f"{doe_id}.csv"

        self.mode = mode
        if mode == MODE.AUTO:
            if self.path.is_file():
                self.sample = _sample_load(self.path, self.id, problem, size)
            else:
                self.sample = _sample_generate(self.id, problem, doe, options)
                _sample_save(self.path, self.sample)
        elif mode == MODE.LOAD:
            self.sample = _sample_load(self.path, self.id, problem, size)
        elif mode == MODE.GENERATE:
            self.sample = _sample_generate(self.id, problem, doe, options)
        else:
            raise ValueError(f"Unknown mode {mode}")

    @property
    def xf(self):
        return self.sample.x, self.sample.f

    @property
    def dim_x(self):
        return self.problem.dim_x

    @property
    def dim_f(self):
        return self.problem.dim_f

    def __repr__(self):
        return self.id

    def __str__(self):
        return self.id

    @property
    def features(self) -> Dict[str, float]:
        path = _features_path(self.path)
        if self.mode == MODE.AUTO:
            if path.is_file():
                return _features_load(path)
            else:
                features = self.sample.features()
                _features_save(path, features)
                return features
        elif self.mode == MODE.LOAD:
            return _features_load(path)
        elif self.mode == MODE.GENERATE:
            return self.sample.features()

    @property
    def models_by_metric(self) -> Dict[METRICS, List[ModelInfo]]:
        path = _models_path(self.path)
        if self.mode == MODE.AUTO:
            if path.is_file():
                return _models_load(path)
            else:
                models_by_metric = self.sample.models_by_metric()
                _models_save(path, models_by_metric)
                return models_by_metric
        elif self.mode == MODE.LOAD:
            return _models_load(path)
        elif self.mode == MODE.GENERATE:
            return self.sample.models_by_metric()


def _sample_save(path: Path, sample: Sample) -> None:
    LOG.debug(f"Saving {path}")
    data = np.hstack((sample.x, sample.f))
    header = ",".join(sample.problem.variable_names + sample.problem.objective_names)
    path.parent.mkdir(parents=True, exist_ok=True)
    np.savetxt(path, data, header=header, delimiter=",", comments="")


def _sample_load(path: Path, id: str, problem: ProblemBase, size: int) -> Sample:
    if not path.is_file():
        raise Exception("Can not find the file " + path)
    LOG.debug(f"Loading {path}")
    data = np.loadtxt(path, dtype=float, skiprows=1, delimiter=",", ndmin=2)
    shape = (size, problem.dim_x + problem.dim_f)
    assert data.shape == shape, "Invalid shape of sample in file"
    return Sample(
        id=id,
        problem=problem,
        x=data[:, 0 : problem.dim_x],
        f=data[:, problem.dim_x : problem.dim_x + problem.dim_f],
    )


def _sample_generate(id: str, problem: ProblemBase, doe: DOE, options: dict) -> Sample:
    LOG.debug(f"Generating {id} {options}")
    x, f = getattr(problem, doe)(**options)
    return Sample(id=id, problem=problem, x=x, f=f)


def _features_path(sample_path: Path, feature: Optional[SAMPLE_FEATURE] = None) -> Path:
    return Path(
        f"{sample_path}.features"
        if feature is None
        else f"{sample_path}.features.{feature}"
    )


def _features_estimate(
    id: str,
    x: np.ndarray,
    f: np.ndarray,
    bounds: Tuple[List[float], List[float]],
    feature: SAMPLE_FEATURE,
    *args,
    **kwargs,
) -> dict:
    from .ela import analyze_sample

    if feature not in SAMPLE_FEATURE:
        raise ValueError(f"Features `{feature}` are not available for sample")
    LOG.debug(f"Estimating `{feature}` features of sample {id}")
    f = f.flatten()
    if SAMPLE_FEATURE.bounds_required(feature):
        lb, ub = bounds
        # CM features check for list bound and fail with tuples
        lb, ub = list(lb), list(ub)
        return analyze_sample(feature, x, f, lb, ub, *args, **kwargs)
    else:
        return analyze_sample(feature, x, f, *args, **kwargs)


def _features_save(path: Path, features: dict) -> None:
    LOG.debug(f"Saving {path}")
    with path.open("w") as file:
        json.dump(features, file, indent=4)


def _features_load(path: Path):
    if not path.is_file():
        raise Exception("Can not find the file " + path)
    LOG.debug(f"Loading {path}")
    with path.open("r") as file:
        return json.load(file)


def _models_path(sample_path: str, metric: Optional[METRICS] = None) -> Path:
    return Path(
        f"{sample_path}.models" if metric is None else f"{sample_path}.model.{metric}"
    )


def _models_estimate(
    id: str,
    problem: ProblemBase,
    x: np.ndarray,
    f: np.ndarray,
    metric: METRICS,
) -> List[ModelInfo]:
    from .model import models_estimate

    if metric not in METRICS:
        raise ValueError(f"Unknown best model metric `{metric}`")
    LOG.debug(f"Finding best model of `{metric}` metric for sample {id}")
    # Note that we assume maximization problem
    models, metrics = models_estimate(problem, metric, x, -f)
    return [
        ModelInfo(
            nu=model.kernel_.nu,
            length_scale=model.kernel_.length_scale,
            metric_value=metric_value,
        )
        for model, metric_value in zip(models, metrics)
    ]


def _models_save(path: Path, models_by_metric: Dict[METRICS, List[ModelInfo]]) -> None:
    LOG.debug(f"Saving {path}")
    models_json = {}
    for metric, models in models_by_metric.items():
        models_json[metric.value] = [asdict(m) for m in models]
    with path.open("w") as file:
        json.dump(models_json, file, indent=4)


def _models_load(path: Path) -> Dict[METRICS, List[ModelInfo]]:
    if not path.is_file():
        raise Exception("Can not find the file " + path)
    LOG.debug(f"Loading {path}")
    with path.open("r") as file:
        models_json = json.load(file)
        models_by_metric = {}
        for metric, models in models_json.items():
            models_by_metric[METRICS(metric)] = [ModelInfo(**m) for m in models]
        return models_by_metric



import ioh

from . import _problem


class Problem(_problem.ProblemBBOB):
    def __init__(self, dim_x=2, instance=0):
        super().__init__(
            ioh.get_problem(
                fid="StepEllipsoid",
                instance=instance,
                dimension=dim_x,
                problem_class=ioh.ProblemClass.BBOB,
            )
        )



from ..functions import bukin02
from ..problem import ProblemBase


class Problem(ProblemBase):
    def init(self):
        self.add_variable(bounds=(-15, -5), initial_guess=-6)
        self.add_variable(bounds=(-3, 3), initial_guess=2)
        self.add_objective()
        self.add_solution(x=[-15.0, 0.0], f=-124.75)

    def calc_objectives(self, x):
        return [bukin02(*x)]



from ..functions import gulf
from ..problem import ProblemBase


class Problem(ProblemBase):
    def init(self):
        self.add_variable(bounds=(0.1, 100), initial_guess=80)
        self.add_variable(bounds=(0.1, 100), initial_guess=80)
        self.add_variable(bounds=(0.1, 100), initial_guess=80)
        self.add_objective()
        self.add_solution(x=[50, 25, 1.5], f=0)

    def calc_objectives(self, x):
        return [gulf(*x)]



from pathlib import Path

_all = []
__all__ = [
    name.stem
    for name in Path(__file__).parent.iterdir()
    if name.suffix == ".py" and not name.stem.startswith("_")
]

for _name in __all__:
    exec(f"from . import {_name}")
    exec(f"_all.append({_name})")

# Sort by name
_all = sorted(_all, key=lambda problem: problem.Problem().NAME)



import logging
import warnings
from dataclasses import asdict
from typing import Callable, Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from bayes_opt.util import UtilityFunction, acq_max
from matplotlib import cm
from sklearn.gaussian_process import GaussianProcessRegressor, kernels

from .constants import METRICS, METRICS_MAX, METRICS_MIN
from .ela import vm_features, vm_triples
from .problems.problem import ProblemBase
from .sample import DOE, SampleCached

LOG = logging.getLogger(__name__)


def build_gp_matern(
    x: np.ndarray,
    f: np.ndarray,
    nu: float = 1.5,
    length_scale: float = 1.0,
    length_scale_bounds: Union[str, Tuple[float, float]] = (1e-5, 1e5),
    n_restarts_optimizer: int = 0,
    seed: int = 0,
):
    kernel = kernels.Matern(
        length_scale=length_scale,
        length_scale_bounds=length_scale_bounds,
        nu=nu,
    )
    model = GaussianProcessRegressor(
        kernel=kernel,
        alpha=1e-6,
        normalize_y=True,
        n_restarts_optimizer=n_restarts_optimizer,
        random_state=seed,
    )
    # Sklearn's GP throws a large number of warnings at times, but
    # we don't really need to see them here.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        model.fit(x, f)
    return model


def calc_suitability(
    model: GaussianProcessRegressor,
    x: np.ndarray,
    f: np.ndarray,
) -> float:
    f_model = model.predict(x)
    if np.isnan(f_model).all():
        # Failed to build model
        return -np.inf, -np.inf
    idx = np.triu_indices(len(x), 1)
    f_diffs = f.reshape(-1, 1) - f.flatten()
    f_diffs_model = f_model.reshape(-1, 1) - f_model
    f_diff_signs = np.sign(np.array(f_diffs[idx]))
    f_diff_signs_model = np.sign(f_diffs_model[idx])
    return np.mean(f_diff_signs == f_diff_signs_model)


def calc_r2_cv(
    model: GaussianProcessRegressor,
    x: np.ndarray,
    f: np.ndarray,
    cv: int = 5,
) -> float:
    from sklearn.model_selection import cross_val_score

    # Sklearn's GP throws a large number of warnings at times, but
    # we don't really need to see them here.
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return cross_val_score(model, x, f, cv=cv).mean()


def calc_suggested_max_f(
    problem: ProblemBase,
    model: GaussianProcessRegressor,
    f_max: float,
    utility_function: UtilityFunction = None,
) -> float:
    # The model should be built for maximization, i.e. with -f values
    if utility_function is None:
        utility_function = UtilityFunction()
    suggestion_x = acq_max(
        ac=utility_function.utility,
        gp=model,
        y_max=f_max,
        bounds=np.array([_ for _ in problem.variable_bounds]).T,
        random_state=np.random.RandomState(0),
    )
    # We assume maximization problem, so return negative of real problem value
    return float(-problem.calc(suggestion_x))


def models_estimate(
    problem: ProblemBase,
    metric: METRICS,
    x: np.ndarray,
    f: np.ndarray,
) -> Tuple[List[GaussianProcessRegressor], List[float]]:
    if metric in {METRICS.SUIT_EXT, METRICS.VM_ANGLES_EXT, METRICS.SEGMENT_EXT}:
        triples = vm_triples.collect_space_filling(x, 2 * x.size, symmetric=True)
        vm = vm_features.VariabilityMap(x, f, triples)
        vm_ext = vm.copy_extended(n_splits_min=2, n_splits_max=2)
        vm_ext_seg = vm_features.Segmentation(vm_ext)

    models: List[GaussianProcessRegressor] = []
    metric_values: List[float] = []
    for nu in sorted([0.5, 1.5, 2.5, np.inf] + [1, 2, 3]):
        model = build_gp_matern(x, f, nu=nu, n_restarts_optimizer=5)
        models.append(model)
        if np.isnan(model.predict(x)).all():
            metric_values.append(np.nan)
            continue

        if metric == METRICS.R2_CV:
            metric_value = calc_r2_cv(model, x, f)
        # elif metric == METRICS.SUGGESTED_F:
        #     metric_value = calc_suggested_max_f(problem, model, f.max())
        elif metric == METRICS.SUIT_EXT:
            metric_value = calc_suitability(model, vm_ext.x, vm_ext.f)
        elif metric == METRICS.VM_ANGLES_EXT:
            metric_value = vm_ext.validate(model)
        elif metric == METRICS.SEGMENT_EXT:
            metric_value = vm_ext_seg.validate(model, n_sectors=19)
        else:
            raise ValueError(f"Unknown model selection metric {metric}")

        metric_values.append(metric_value)
    return models, metric_values


def models_select(metric_values: List[float], metric: METRICS) -> int:
    if len(metric_values) == 1:
        return 0

    best_value = None
    best_index = None
    for i, metric_value in enumerate(metric_values):
        if not np.isfinite(metric_value):
            continue

        if best_value is None:
            best_value = metric_value
            best_index = i
        elif metric in METRICS_MIN:
            if metric_value < best_value:
                best_value = metric_value
                best_index = i
        elif metric in METRICS_MAX:
            if metric_value > best_value:
                best_value = metric_value
                best_index = i
        else:
            raise ValueError(f"Can not compare models for metric {metric}")

    assert best_value is not None
    assert best_index is not None
    return best_index


def analyze_sample_features(
    problem_modules: List,
    dim: int,
    sizes: List[int],
    metric: METRICS,
    seed: int,
) -> pd.DataFrame:
    data_rows = []
    for problem_module in problem_modules:
        for size in sizes:
            problem: ProblemBase = problem_module.Problem(dim)
            sample = SampleCached(
                problem,
                size=size,
                doe=DOE.RAND,
                options={"seed": seed},
            )
            sample_features = sample.features
            models_info = sample.models_by_metric[metric]
            best_i = models_select([m.metric_value for m in models_info], metric)
            best_model_info = asdict(models_info[best_i])
            data_rows.append(
                pd.Series(
                    {
                        # Problem definition
                        "problem_id": problem.id,
                        "problem": problem.NAME,
                        "dim": dim,
                        "size": size,
                        "seed": seed,
                        "metric": metric,
                        **sample_features,
                        **best_model_info,
                    }
                )
            )
    return pd.DataFrame(data_rows)


def plot_surface(
    func: Callable[[np.ndarray], np.ndarray],
    lower_bound: Tuple[float, float],
    upper_bound: Tuple[float, float],
    n_nodes_x1: int = 50,
    n_nodes_x2: int = 50,
    n_nodes_f: int = 50,
    contour: bool = False,
    wireframe: bool = False,
    file_name: bool = None,
    title: Optional[str] = None,
    size: Tuple[float, float] = (6, 5),
    dpi: int = 100,
    sample_x: np.ndarray = None,
    sample_f: np.ndarray = None,
    samples: Dict[str, Tuple[np.ndarray, np.ndarray]] = {},
    samples_m: Dict[str, str] = {},  # marker
    samples_c: Dict[str, str] = {},  # color
):
    lb1, lb2 = lower_bound[:2]
    ub1, ub2 = upper_bound[:2]
    x1_nodes = np.linspace(lb1, ub1, n_nodes_x1)
    x2_nodes = np.linspace(lb2, ub2, n_nodes_x2)
    x1_grid, x2_grid = np.meshgrid(x1_nodes, x2_nodes)
    x_grid = np.vstack([x1_grid.flatten(), x2_grid.flatten()]).T
    y_grid = func(x_grid)
    y_grid = np.reshape(y_grid, (x1_grid.shape[0], -1))
    if contour:
        plt.figure(figsize=size, dpi=dpi)
        plt.contourf(x1_grid, x2_grid, y_grid, cmap=cm.coolwarm, levels=n_nodes_f)
        plt.colorbar()
        plt.xlabel("$x_1$")
        plt.ylabel("$x_2$")
        if sample_x is not None:
            sample_x = np.atleast_2d(sample_x)
            plt.plot(sample_x[:, 0], sample_x[:, 1], ".", c="k")
            if sample_f is not None:
                [
                    plt.text(xi[0], xi[1], f"{i}|{fi[0]:.1f}", fontsize=8)
                    for i, (xi, fi) in enumerate(zip(sample_x, sample_f))
                ]
        for label, (x, f) in samples.items():
            x = np.atleast_2d(x)
            plt.plot(
                x[:, 0],
                x[:, 1],
                samples_m.get(label, "."),
                c=samples_c.get(label, None),
                label=label,
            )
            if f is not None:
                [
                    plt.text(xi[0], xi[1], f"{i}|{fi[0]:.1f}", fontsize=8)
                    for i, (xi, fi) in enumerate(zip(x, f))
                ]
    else:
        fig = plt.figure(figsize=size, dpi=dpi)
        ax = fig.add_subplot(projection="3d")
        ax.set_box_aspect(aspect=None, zoom=0.8)
        ax.set_xlabel("$x_1$")
        ax.set_ylabel("$x_2$")
        ax.set_zlabel("$f$")
        if wireframe:
            ax.plot_wireframe(x1_grid, x2_grid, y_grid)
        else:
            ax.plot_surface(x1_grid, x2_grid, y_grid, cmap=cm.coolwarm)
        if sample_x is not None and sample_f is not None:
            sample_x = np.atleast_2d(sample_x)
            sample_f = np.atleast_2d(sample_f)
            ax.scatter3D(sample_x[:, 0], sample_x[:, 1], sample_f[:, 0], ".", c="k")
        for label, (x, f) in samples.items():
            x = np.atleast_2d(x)
            f = np.atleast_2d(f)
            ax.scatter3D(
                x[:, 0],
                x[:, 1],
                f[:, 1],
                samples_m.get(label, "."),
                c=samples_c.get(label, None),
                label=label,
            )
    plt.grid()
    if samples:
        plt.legend()
    if title is not None:
        plt.title(title)
    if file_name is None:
        plt.show()
    else:
        plt.savefig(file_name)
        plt.close()



import ioh

from ..problem import ProblemBase


class ProblemBBOB(ProblemBase):
    def __init__(self, ioh_problem: ioh.problem.RealSingleObjective):
        self.ioh_problem = ioh_problem

    def get_instance(self, instance):
        id = self.ioh_problem.meta_data.problem_id
        dim = self.ioh_problem.meta_data.n_variables
        problem = ProblemBBOB(ioh.get_problem(id, instance, dim))
        problem.NAME = self.NAME
        return problem

    def init(self):
        for lb, ub in zip(self.ioh_problem.bounds.lb, self.ioh_problem.bounds.ub):
            self.add_variable(bounds=(lb, ub), initial_guess=None)
        self.add_objective()
        self.add_solution(x=self.ioh_problem.optimum.x, f=self.ioh_problem.optimum.y)

    def calc_objectives(self, x):
        return [self.ioh_problem(x)]


def generate():
    from pathlib import Path

    def to_snake(name):
        return "".join(
            [
                "_" + c.lower() if i > 0 and c.isupper() else c.lower()
                for i, c in enumerate(name)
            ]
        )

    for i, name in ioh.ProblemClass.BBOB.problems.items():
        path = Path(__file__).parent / f"{to_snake(name)}.py"
        path.touch(exist_ok=True)
        print(f"Updating class {path}")
        with path.open("w") as f:
            f.write(
                f"""import ioh

from . import _problem


class Problem(_problem.ProblemBBOB):
    def __init__(self, dim_x=2, instance=0):
        super().__init__(
            ioh.get_problem(
                fid="{name}",
                instance=instance,
                dimension=dim_x,
                problem_class=ioh.ProblemClass.BBOB,
            )
        )
"""
            )



from ..functions import keane
from ..problem import ProblemBase


class Problem(ProblemBase):
    def init(self):
        self.add_variable(bounds=(-0.01, 10), initial_guess=10)
        self.add_variable(bounds=(-0.01, 10), initial_guess=10)
        self.add_objective()
        self.add_solution(x=[0.0, 1.39325], f=-0.6737)

    def calc_objectives(self, x):
        return [keane(*x)]



import numpy as np

from ..functions import branin
from ..problem import ProblemBase


class Problem(ProblemBase):
    def init(self):
        self.add_variable(bounds=(-5, 10), initial_guess=5)
        self.add_variable(bounds=(0, 15), initial_guess=15)
        self.add_objective()
        self.add_solution(x=[-np.pi, 12.275], f=0.3978873577)
        self.add_solution(x=[np.pi, 2.275], f=0.3978873577)
        self.add_solution(x=[9.42478, 2.475], f=0.3978873577)

    def calc_objectives(self, x):
        return [branin(*x)]



from ..functions import brent
from ..problem import ProblemBase


class Problem(ProblemBase):
    def init(self):
        self.add_variable(bounds=(-100, 100), initial_guess=80)
        self.add_variable(bounds=(-100, 100), initial_guess=80)
        self.add_objective()
        self.add_solution(x=[-10.0, -10.0], f=0)

    def calc_objectives(self, x):
        return [brent(*x)]



from ..functions import elatar
from ..problem import ProblemBase


class Problem(ProblemBase):
    def init(self):
        self.add_variable(bounds=(-100, 100), initial_guess=80)
        self.add_variable(bounds=(-100, 100), initial_guess=80)
        self.add_objective()
        self.add_solution(x=[3.409186, -2.171433], f=1.7127803540)

    def calc_objectives(self, x):
        return [elatar(*x)]



from ..functions import deckkersaarts
from ..problem import ProblemBase


class Problem(ProblemBase):
    def init(self):
        self.add_variable(bounds=(-20, 20), initial_guess=18)
        self.add_variable(bounds=(-20, 20), initial_guess=18)
        self.add_objective()
        self.add_solution(x=[0, 15], f=-24777)

    def calc_objectives(self, x):
        return [deckkersaarts(*x)]



from ..functions import fon1, fon2
from ..problem import ProblemBase, Solution


class Problem(ProblemBase):
    def __init__(self, dimension=None):
        self.dimension = dimension or 2

    def init(self):
        for _ in range(self.dimension):
            self.add_variable(bounds=(-4, 4), initial_guess=4)
        self.add_objective()
        self.add_objective()

    def calc_objectives(self, x):
        return [fon1(x), fon2(x)]



from ..functions import deceptive
from ..problem import ProblemBase


class Problem(ProblemBase):
    def __init__(self, dim_x=2):
        self.__dim_x = dim_x
        self._alpha = 0.5

    def init(self):
        for _ in range(self.__dim_x):
            self.add_variable(bounds=(0, 1), initial_guess=0.8)
        self.add_objective()
        self.add_solution(x=[self._alpha] * self.__dim_x, f=-1)

    def calc_objectives(self, x):
        return [deceptive(x, self._alpha)]



from ..functions import beale
from ..problem import ProblemBase


class Problem(ProblemBase):
    def init(self):
        self.add_variable(bounds=(-10, 10), initial_guess=8)
        self.add_variable(bounds=(-10, 10), initial_guess=8)
        self.add_objective()
        self.add_solution(x=[3.0, 0.5], f=0)

    def calc_objectives(self, x):
        return [beale(*x)]



from ..functions import trid
from ..problem import ProblemBase


class Problem(ProblemBase):
    def __init__(self, dim_x=2):
        self.__dim_x = dim_x

    def init(self):
        for _ in range(self.__dim_x):
            self.add_variable(
                bounds=(-self.__dim_x**2, self.__dim_x**2),
                initial_guess=-self.__dim_x**2,
            )
        self.add_objective()
        self.add_solution(
            x=[(i + 1) * (self.__dim_x - i) for i in range(self.__dim_x)],
            f=-self.__dim_x * (self.__dim_x + 4) * (self.__dim_x - 1) / 6.0,
        )

    def calc_objectives(self, x):
        return [trid(x)]



from ..functions import wolfe
from ..problem import ProblemBase


class Problem(ProblemBase):
    def init(self):
        self.add_variable(bounds=(0, 2), initial_guess=2)
        self.add_variable(bounds=(0, 2), initial_guess=2)
        self.add_variable(bounds=(0, 2), initial_guess=2)
        self.add_objective()
        self.add_solution(x=[0, 0, 0], f=0)

    def calc_objectives(self, x):
        return [wolfe(*x)]



from ..functions import amgm
from ..problem import ProblemBase


class Problem(ProblemBase):
    def __init__(self, dim_x=2):
        self.__dim_x = dim_x

    def init(self):
        for i in range(self.__dim_x):
            self.add_variable(bounds=(0, 10), initial_guess=i % 10)
        self.add_objective()
        self.add_solution(x=[0.0] * self.__dim_x, f=0)

    def calc_objectives(self, x):
        return [amgm(x)]



import numpy as np
import pandas as pd


def features_icp(x, f, triples, n=16):
    vm_ic = ela.vmap.VMIC(x, f, triples)
    eps, ic, icp = vm_ic.run_all_1d(np.linspace(vm_ic._eps_min, vm_ic.eps_max, n // 2))
    return ic, icp


def features_seg(x, f, triples, n=None):
    vm = ela.VMSegment(x, f, triples)
    if n is None:  # 7 output features
        gamma, seg_p = vm.run_all(sectors=[0, 30, 45, 60, 90, 120, 135, 150, 180])
    else:
        gamma, seg_p = vm.run_all(n_sectors=n + 2)
    return seg_p[1:-1].tolist()


def features_gic(x, f, triples, n=16):
    vm_ic = ela.vmap.VMIC(x, f, triples)
    eps_values = np.linspace(vm_ic._eps_min, vm_ic.eps_max, int(np.sqrt(n)))
    return vm_ic.run_all_2d(eps_values, eps_values)[1]


def features_ic(x, f, triples, n=8):
    vm_ic = ela.vmap.VMIC(x, f, triples)
    eps_values = np.linspace(vm_ic._eps_min, vm_ic.eps_max, n // 2)
    ic_max_eps1 = [
        vm_ic.run_all_2d(eps_i, np.geomspace(vm_ic._eps_min, vm_ic.eps_max, 100))[
            1
        ].max()
        for eps_i in eps_values
    ]
    ic_max_eps2 = [
        vm_ic.run_all_2d(np.geomspace(vm_ic._eps_min, vm_ic.eps_max, 100), eps_i)[
            1
        ].max()
        for eps_i in eps_values
    ]
    return np.hstack((ic_max_eps1, ic_max_eps2)).tolist()


def features_segic(x, f, triples):
    # return features_seg(x, f, triples, n=11), features_ic(x, f, triples, n=16) # features_old analogue
    return features_seg(x, f, triples, n=None), features_ic(x, f, triples, n=8)


def features_old(x, f, triples):
    vm = ela.VMSegment(x, f, triples)
    features1 = vm.run_all(13)[1][1:-1]

    vm_ic = ela.vmap.VMIC(x, f, triples)
    eps_values = np.linspace(vm_ic._eps_min, vm_ic.eps_max, 8)
    ic1 = [
        vm_ic.run_all_2d(e, np.geomspace(vm_ic._eps_min, vm_ic.eps_max, 100))[1].max()
        for e in eps_values
    ]
    ic2 = [
        vm_ic.run_all_2d(np.geomspace(vm_ic._eps_min, vm_ic.eps_max, 100), e)[1].max()
        for e in eps_values
    ]
    features2 = np.hstack((ic1, ic2))

    return features1.tolist(), features2.tolist()



from ..functions import sixhumpcamel
from ..problem import ProblemBase


class Problem(ProblemBase):
    def init(self):
        self.add_variable(bounds=(-3, 3), initial_guess=3)
        self.add_variable(bounds=(-2, 2), initial_guess=2)
        self.add_objective()
        self.add_solution(x=[0.0898, -0.7126], f=-1.03164)
        self.add_solution(x=[-0.0898, 0.7126], f=-1.03164)

    def calc_objectives(self, x):
        return [sixhumpcamel(*x)]



from ..functions import dixonprice
from ..problem import ProblemBase


class Problem(ProblemBase):
    def __init__(self, dim_x=2):
        self.__dim_x = dim_x

    def init(self):
        for _ in range(self.__dim_x):
            self.add_variable(bounds=(-10, 10), initial_guess=8)
        self.add_objective()
        self.add_solution(
            x=[
                2.0 ** ((2.0 - 2.0 ** (i)) / 2.0 ** (i))
                for i in range(1, self.__dim_x + 1)
            ],
            f=0,
        )

    def calc_objectives(self, x):
        return [dixonprice(x)]



from ..functions import shubert
from ..problem import ProblemBase


class Problem(ProblemBase):
    def __init__(self, dim_x=2):
        self.__dim_x = dim_x

    def init(self):
        for _ in range(self.__dim_x):
            self.add_variable(bounds=(-5.12, 5.12), initial_guess=3)
        self.add_objective()

    def calc_objectives(self, x):
        return [shubert(x)]



from dataclasses import dataclass
from typing import List, Optional, Self, Tuple

import numpy as np

from . import vm_triples


@dataclass
class VariabilityMap:
    x: np.ndarray
    f: np.ndarray
    triples: List[Tuple[int, int, int]]
    dx_norm: Optional[np.ndarray] = None
    deltas: Optional[np.ndarray] = None
    f_stats: Optional[Tuple[float, float]] = None

    def __post_init__(self):
        if self.dx_norm is None:
            self.dx_norm = vm_triples._calc_dx_norm(self.x, self.triples)
        if self.f_stats is None:
            self.f_stats = (self.f.mean(), self.f.std())
            if self.f_stats[1] == 0:
                self.f_stats[1] = 1
        if self.deltas is None:
            f = (self.f.flatten() - self.f_stats[0]) / self.f_stats[1]
            self.deltas = np.diff(f[self.triples], axis=1) / self.dx_norm
        assert np.all(self.dx_norm > 0)

    def copy_extended(self, **kwargs) -> Self:
        x, f, triples = vm_triples.extend(self.x, self.f, self.triples, **kwargs)
        return VariabilityMap(x=x, f=f, triples=triples, f_stats=self.f_stats)

    def copy_with_model(self, model):
        return VariabilityMap(
            x=self.x,
            f=model.predict(self.x),
            triples=self.triples,
            dx_norm=self.dx_norm,
            f_stats=self.f_stats,
        )

    def validate(self, model) -> float:
        vm_model = self.copy_with_model(model)
        # TODO how we validate triples that have both 0 deltas, which means that points
        # are at constant region? Now such angles are just nan
        delta_angles = vm_triples._calc_angle_vectors(self.deltas, vm_model.deltas)
        return np.nanmean(delta_angles)

    def plot(
        self,
        scale="linear",
        fontsize=12,
        title: str = "",
        labels: np.ndarray = None,
        size: Tuple[int, int] = (8, 8),
        dpi: int = 100,
        file_name: Optional[str] = None,
    ):
        import matplotlib.pyplot as plt

        plt.figure(figsize=size, dpi=dpi)

        plt.plot(self.deltas[:, 0], self.deltas[:, 1], ".")
        if labels is not None:
            [plt.text(d0, d1, f"{l:.1f}") for (d0, d1), l in zip(self.deltas, labels)]
        plt.xlabel("$\delta_1$", fontsize=fontsize)
        plt.ylabel("$\delta_2$", fontsize=fontsize)
        plt.xscale(scale)
        plt.yscale(scale)
        x_lim = np.abs(plt.xlim()).max()
        y_lim = np.abs(plt.ylim()).max()
        plt.plot([-x_lim, x_lim], [y_lim, -y_lim], "-", c="k", alpha=0.5)
        plt.xlim([-x_lim, x_lim])
        plt.ylim([-y_lim, y_lim])
        plt.grid()
        plt.title(title)
        if file_name:
            plt.savefig(file_name)
            plt.close()
        else:
            plt.show()

    def plot_model(
        self,
        model,
        scale="linear",
        fontsize=12,
        title: str = "",
        size: Tuple[int, int] = (8, 8),
        dpi: int = 100,
        file_name: Optional[str] = None,
        model_tag: str = "",
    ):
        import matplotlib.pyplot as plt

        plt.figure(figsize=size, dpi=dpi)

        vm_model = self.copy_with_model(model)
        plt.plot(self.deltas[:, 0], self.deltas[:, 1], ".", c="k", label="sample")
        plt.plot(
            vm_model.deltas[:, 0],
            vm_model.deltas[:, 1],
            ".",
            alpha=0.7,
            markersize=2,
            label=model_tag or "model",
        )
        plt.xlabel("$\delta_1$", fontsize=fontsize)
        plt.ylabel("$\delta_2$", fontsize=fontsize)
        plt.xscale(scale)
        plt.yscale(scale)
        x_lim = np.abs(plt.xlim()).max()
        y_lim = np.abs(plt.ylim()).max()
        plt.plot([-x_lim, x_lim], [y_lim, -y_lim], "-", c="k", alpha=0.5)
        plt.xlim([-x_lim, x_lim])
        plt.ylim([-y_lim, y_lim])
        for delta, delta_step in zip(self.deltas, vm_model.deltas - self.deltas):
            plt.arrow(delta[0], delta[1], delta_step[0], delta_step[1], alpha=0.2)
        plt.legend()
        plt.grid()
        plt.title(title)
        if file_name:
            plt.savefig(file_name)
            plt.close()
        else:
            plt.show()


class Segmentation:
    def __init__(self, vm: VariabilityMap):
        self.vm = vm
        d0, d1 = vm.deltas[:, 0], vm.deltas[:, 1]
        self.angles: np.ndarray = np.angle(d0 + d1 * 1j, deg=True)
        # Make 0 angle at -45 degrees and +-180 angle at 145 degrees. That way it's
        # easier to generate sectors with respect to symmetry line of variability map.
        self.angles += 45
        self.angles[self.angles > 180] -= 360
        # Weight of triple is set by angle between its points:
        # - weight=0 if angle is 0 degrees
        # - weight=1if angle is 180 degrees
        self.weights = vm_triples._calc_angle(vm.x, vm.triples) / 180.0
        self.weights_sum = sum(self.weights)

    def _mask_sector(self, gamma: float) -> np.ndarray:
        assert 0 <= gamma <= 180
        angles_abs = np.abs(self.angles)
        return angles_abs <= gamma

    def _mask_central_sector(self, gamma: float) -> np.ndarray:
        assert 0 <= gamma <= 180
        angle_min = 90 - gamma / 2.0
        angle_max = 90 + gamma / 2.0
        angles_abs = np.abs(self.angles)
        return (angle_min <= angles_abs) * (angles_abs <= angle_max)

    def sector_ratio(self, gamma):
        mask = self._mask_sector(gamma)
        return sum(self.weights[mask]) / self.weights_sum

    def sector_ratios_grid(self, n_sectors=181):
        gammas = np.linspace(0, 180, n_sectors)
        ratios = np.zeros(gammas.size)
        for i, gamma in enumerate(gammas):
            mask = self._mask_sector(gamma)
            ratios[i] = sum(self.weights[mask])
        return gammas, ratios / self.weights_sum

    def central_sector_ratio(self, gamma):
        mask = self._mask_central_sector(gamma)
        return sum(self.weights[mask]) / self.weights_sum

    def central_sector_ratios_grid(self, n_sectors=181):
        gammas = np.linspace(0, 180, n_sectors)
        ratios = np.zeros(gammas.size)
        for i, gamma in enumerate(gammas):
            mask = self._mask_central_sector(gamma)
            ratios[i] = sum(self.weights[mask])
        return gammas, ratios / self.weights_sum

    def validate(self, model, n_sectors=19) -> float:
        sample_ratios = self.sector_ratios_grid(n_sectors)[1]
        vm_model = self.vm.copy_with_model(model)
        model_ratios = Segmentation(vm_model).sector_ratios_grid(n_sectors)[1]
        return np.nanmean(np.abs(sample_ratios - model_ratios))

    def validate_central(self, model, n_sectors=19):
        sample_ratios = self.central_sector_ratios_grid(n_sectors)[1]
        vm_model = self.vm.copy_with_model(model)
        model_ratios = Segmentation(vm_model).central_sector_ratios_grid(n_sectors)[1]
        return np.nanmean(np.abs(sample_ratios - model_ratios))

    def plot(self, model=None, n_sectors=181):
        import matplotlib.pyplot as plt

        gammas, ratios = self.sector_ratios_grid(n_sectors)
        plt.plot(gammas, ratios, label="sample")
        if model is not None:
            vm_model = self.vm.copy_with_model(model)
            vm_seg = Segmentation(vm_model)
            gammas_m, ratios_m = vm_seg.sector_ratios_grid(n_sectors)
            plt.plot(gammas_m, ratios_m, label="model")
        plt.axvline(x=0, color="k", linestyle="--", alpha=0.4)
        plt.axvline(x=45, color="k", linestyle="--", alpha=0.4)
        plt.axvline(x=90, color="k", linestyle="--", alpha=0.4)
        plt.axvline(x=135, color="k", linestyle="--", alpha=0.4)
        plt.axvline(x=180, color="k", linestyle="--", alpha=0.4)
        plt.xticks(range(0, 181, 45))
        plt.legend()
        plt.grid()
        plt.show()

    def plot_central(self, model=None, n_sectors=181):
        import matplotlib.pyplot as plt

        gammas, ratios = self.central_sector_ratios_grid(n_sectors)
        plt.plot(gammas, ratios, label="sample")
        if model is not None:
            vm_model = self.vm.copy_with_model(model)
            vm_seg = Segmentation(vm_model)
            gammas_m, ratios_m = vm_seg.central_sector_ratios_grid(n_sectors)
            plt.plot(gammas_m, ratios_m, label="model")
        plt.axvline(x=0, color="k", linestyle="--", alpha=0.4)
        plt.axvline(x=45, color="k", linestyle="--", alpha=0.4)
        plt.axvline(x=90, color="k", linestyle="--", alpha=0.4)
        plt.axvline(x=135, color="k", linestyle="--", alpha=0.4)
        plt.axvline(x=180, color="k", linestyle="--", alpha=0.4)
        plt.xticks(range(0, 181, 45))
        plt.legend()
        plt.grid()
        plt.show()


class InformationContent:
    """
    +--------+--------+--------+--------+
    |   i\j  | [=]  0 | [-] -1 | [+]  2 |
    +--------+--------+--------+--------+
    | [=]  0 |    0   |   -1   |    2   |
    +--------+--------+--------+--------+
    | [-] -1 |    1   |    0   |    3   |
    +--------+--------+--------+--------+
    | [+]  2 |   -2   |   -3   |    0   |
    +--------+--------+--------+--------+
    """

    _STEP = {"negative": -1, "neutral": 0, "positive": 2}

    def __init__(self, vm: VariabilityMap):
        self.vm = vm
        steps = list(self._STEP.values())
        self.block_types = np.array([j - i for i in steps for j in steps if i != j])
        self.ic_base = np.log(self.block_types.size)
        self.weights = vm_triples._calc_angle(vm.x, vm.triples) / 180.0
        self.weights_sum = sum(self.weights)

    def evaluate(
        self, eps1: float, eps2: Optional[float] = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        sequence = np.zeros_like(self.vm.deltas, int)
        if eps2 is None:
            eps2 = eps1
        sequence.fill(self._STEP["neutral"])
        sequence[self.vm.deltas[:, 0] > eps1, 0] = self._STEP["positive"]
        sequence[self.vm.deltas[:, 0] < -eps2, 0] = self._STEP["negative"]
        sequence[self.vm.deltas[:, 1] > eps2, 1] = self._STEP["positive"]
        sequence[self.vm.deltas[:, 1] < -eps1, 1] = self._STEP["negative"]
        blocks = np.diff(sequence, axis=1).ravel()
        # Calculate information content
        ic = 0
        for block_type in self.block_types:
            prob = np.sum(self.weights[blocks == block_type]) / self.weights_sum
            if prob > 0:
                ic -= prob * np.log(prob) / self.ic_base
        # Ignore neutral changes, track only [+-] and [-+] blocks
        mask = blocks == (self._STEP["positive"] - self._STEP["negative"])
        mask += blocks == (self._STEP["negative"] - self._STEP["positive"])
        # Count without blocks with repeated changes
        icp = np.sum(self.weights[mask]) / (self.weights_sum + 1.0)
        return ic, icp

    def evaluate_grid_1d(
        self, eps: Optional[float] = None
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        if eps is None:
            eps = np.unique(np.abs(self.vm.deltas))
        ic, icp = np.vectorize(self.evaluate)(eps1=eps)
        return eps, ic.astype(float), icp.astype(float)

    def evaluate_grid_2d(
        self, eps1: Optional[float] = None, eps2: Optional[float] = None
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        if eps1 is None:
            eps1 = np.unique(self.vm.deltas[:, 0])
        if eps2 is None:
            eps2 = np.unique(self.vm.deltas[:, 1])
        eps = np.vstack([_.flatten() for _ in np.meshgrid(eps1, eps2)]).T
        ic, icp = np.vectorize(self.evaluate)(eps1=eps[:, 0], eps2=eps[:, 1])
        return eps, ic.reshape(-1, 1).astype(float), icp.reshape(-1, 1).astype(float)

    def validate(self, model) -> Tuple[float, float]:
        vm_model = self.vm.copy_with_model(model)
        eps, ic, icp = self.evaluate_grid_1d()
        eps, ic_m, icp_m = InformationContent(vm_model).evaluate_grid_1d(eps)
        return np.abs(ic - ic_m).mean(), np.abs(icp - icp_m).mean()

    def plot(self, model=None, eps_scale="log"):
        import matplotlib.pyplot as plt

        eps, ic, icp = self.evaluate_grid_1d()
        plt.plot(eps, ic, label="sample ic")
        plt.plot(eps, icp, label="sample icp")
        if model is not None:
            vm_model = self.vm.copy_with_model(model)
            vm_ic = InformationContent(vm_model)
            eps_m, ic_m, icp_m = vm_ic.evaluate_grid_1d()
            plt.plot(eps_m, ic_m, label="model ic")
            plt.plot(eps_m, icp_m, label="model icp")

        plt.xscale(eps_scale)
        plt.legend()
        plt.grid()
        plt.show()


def calculate(x: np.ndarray, f: np.ndarray) -> dict:
    triples = vm_triples.collect_space_filling(x, 2 * x.size, symmetric=True)
    vm = VariabilityMap(x, f, triples)
    vm_seg = Segmentation(vm)
    vm_ic = InformationContent(vm)
    deltas = np.abs(vm.deltas)
    delta_min = deltas[deltas > 0].min()  # Should not be zero for geomspace
    delta_max = deltas.max()

    features = {}

    gammas, ratios = vm_seg.sector_ratios_grid()
    idx = [30, 45, 60, 90, 120, 135, 150]
    thresh = [0.05, 0.25, 0.5, 0.75, 0.95]
    features.update({f"sector_{int(g)}": r for g, r in zip(gammas[idx], ratios[idx])})
    features.update({f"ratio_{r}": gammas[np.abs(ratios - r).argmin()] for r in thresh})

    eps_grid = np.linspace(delta_min, delta_max, 4)
    eps, ic, icp = vm_ic.evaluate_grid_1d(eps_grid)
    features.update({f"ic_{i}": ic[i] for i in range(4)})
    features.update({f"icp_{i}": icp[i] for i in range(4)})
    eps, ic, icp = vm_ic.evaluate_grid_1d()

    eps_icp_half = eps[icp < 0.5 * icp[0]]
    features.update(
        {
            "ic_max": ic.max(),
            "ic_max_eps": eps[ic.argmax()],
            "icp_init": icp[0],
            "eps_ic_s": eps[ic < 0.05][0],
            "eps_icp_half": eps_icp_half[0] if len(eps_icp_half) else np.mean(eps),
            "eps_min": eps.min(),
            "eps_max": eps.max(),
        }
    )
    eps_search = np.geomspace(delta_min, delta_max, 200)
    ic_max_eps1 = [vm_ic.evaluate_grid_2d(e, eps_search)[1].max() for e in eps_grid]
    ic_max_eps2 = [vm_ic.evaluate_grid_2d(eps_search, e)[1].max() for e in eps_grid]
    features.update({f"gic_eps1_{i}": ic_max_eps1[i] for i in range(len(eps_grid))})
    features.update({f"gic_eps2_{i}": ic_max_eps2[i] for i in range(len(eps_grid))})

    icp_max_eps1 = [vm_ic.evaluate_grid_2d(e, eps_search)[2].max() for e in eps_grid]
    icp_max_eps2 = [vm_ic.evaluate_grid_2d(eps_search, e)[2].max() for e in eps_grid]
    features.update({f"gicp_eps1_{i}": icp_max_eps1[i] for i in range(len(eps_grid))})
    features.update({f"gicp_eps2_{i}": icp_max_eps2[i] for i in range(len(eps_grid))})

    return {f"vm.{f}": v for f, v in features.items()}



import time
from dataclasses import dataclass, field
from datetime import timedelta
from typing import Dict, Optional, Self


@dataclass
class Timer:
    """Class to simplify time measuring routine"""

    time: Optional[int] = None
    is_running: bool = False

    def start(self) -> Self:
        if self.is_running:
            self.time = time.time() - self.time
            self.is_running = False
            print(f"WARN: timer was reset at {self.str()}")
        self.time = time.time()
        self.is_running = True
        return self

    def stop(self) -> Self:
        if not self.is_running:
            print(f"WARN: timer was not started")
            return
        self.time = time.time() - self.time
        self.is_running = False
        return self

    def str(self) -> str:
        if self.time is None:
            return f"WARN: timer was not started"
        is_tunning = ""
        time_sec = self.time
        if self.is_running:
            is_tunning = " still running"
            time_sec = time.time() - self.time
        return f"<{timedelta(seconds=time_sec)}> ({time_sec:g} sec){is_tunning}"


class set_timer:
    def __init__(self):
        self.timer = Timer()

    def __enter__(self):
        return self.timer.start()

    def __exit__(self, exc_type, exc_value, traceback):
        self.timer.stop()


@dataclass
class Timers:
    """Named timers"""

    timers: Dict[str, Timer] = field(default_factory=dict)

    def start(self, tag: str = "") -> Timer:
        self.timers.setdefault(tag, Timer()).start()
        return self.timers[tag]

    def stop(self, tag: str = "") -> Timer:
        self.timers.setdefault(tag, Timer()).stop()
        return self.timers[tag]

    def str(self, tag: Optional[str] = None) -> str:
        if tag is None:
            return {tag: timer.str() for tag, timer in self.timers.items()}
        elif tag not in self.timers:
            print(f"WARN: timer `{tag}` was not set")
            return {tag: timer.str() for tag, timer in self.timers.items()}
        else:
            return self.timers[tag].str()


timers = Timers()



import numpy as np


def factorize(dim: int, n_points: int) -> np.ndarray[int]:
    points_per_dim = int(np.ceil(np.power(n_points, 1.0 / dim)))
    return np.repeat(points_per_dim, dim)


def cells_grid(
    dim: int,
    size: int,
    random_shift: float = 0,  # value from 0 to 1 to randomly shift points within cells
) -> np.ndarray[float]:
    n_cells = factorize(dim=dim, n_points=size)
    n_points = np.product(n_cells)
    steps = [1.0 / (2.0 * n) for n in n_cells]
    # Nodes are placed at centers of cells from 0 to 1
    nodes = [np.linspace(s, 1 - s, n) for s, n in zip(steps, n_cells)]
    result = np.vstack([n.flatten() for n in np.meshgrid(*nodes)]).T
    if random_shift > 0:
        assert random_shift < 1
        bound = random_shift * np.array(steps)
        result += np.random.uniform(-bound, bound, result.shape)
    if n_points > size:
        random_idx = np.random.choice(np.arange(n_points), size, replace=False)
        result = result[random_idx]
    return result


def full_factorial(
    dim: int,
    size: int,
    randomized: bool = True,
):
    # Fills from 0 to 1 in regular grid
    n_cells = factorize(dim=dim, n_points=size)
    n_points = np.product(n_cells)
    result = np.empty((n_points, dim), dtype=float)
    repeat_count = 1
    tile_count = n_points
    for i, n_i in enumerate(n_cells):
        tile_count //= n_i
        nodes = np.random.random(n_i) if randomized else np.linspace(0, 1, n_i)
        result[:, i] = np.tile(np.repeat(nodes, repeat_count, axis=0), tile_count)
        repeat_count *= n_i
    if n_points > size:
        random_idx = np.random.choice(np.arange(n_points), size, replace=False)
        result = result[random_idx]
    return result



import itertools
import logging
from typing import Dict, List, Optional, Set, Tuple

import numpy as np

PRECISION = 12
LOG = logging.getLogger(__name__)


def _normalize_x(x: np.ndarray) -> np.ndarray:
    # Normalize values along dimensions
    x = np.array(x, copy=True)
    for i in np.arange(x.shape[1]):
        x[:, i] = (x[:, i] - x[:, i].mean()) / x[:, i].std(ddof=1)
    return x


def _normalize_distances(distances):
    n = len(distances)
    distances_triu = distances[np.triu_indices(n, k=1)]
    distances_triu_idx = np.argsort(distances_triu)
    min_dist = distances_triu.min()
    max_dist = distances_triu.max()
    distances_triu_normed = np.linspace(min_dist, max_dist, len(distances_triu))
    distances_triu_normed = distances_triu_normed[np.argsort(distances_triu_idx)]
    distances_normed = np.zeros_like(distances)
    distances_normed[np.triu_indices(n, k=1)] = distances_triu_normed
    distances_normed += distances_normed.T
    return distances_normed


def _calc_angle_vectors(v1: np.ndarray, v2: np.ndarray) -> np.ndarray:
    assert v1.shape == v2.shape
    # dot product for each row
    dot = np.einsum("ij, ij->i", v1, v2)
    norm = np.linalg.norm(v1, axis=1) * np.linalg.norm(v2, axis=1)
    # May be nan for zero-length vectors
    norm[norm == 0] = np.nan
    cosine_angle = np.round(dot / norm, PRECISION)
    angle = np.arccos(cosine_angle)
    np.degrees(angle, out=angle)
    return angle


def _calc_angle(x: np.ndarray, triples: List[Tuple[int, int, int]]) -> np.ndarray:
    v1 = x[triples[:, 0]] - x[triples[:, 1]]
    v2 = x[triples[:, 2]] - x[triples[:, 1]]
    return _calc_angle_vectors(v1, v2)


def _calc_dx_norm(x: np.ndarray, triples: List[Tuple[int, int, int]]) -> np.ndarray:
    dx_norm = np.hypot.reduce(np.diff(x[triples], axis=1), axis=2)
    np.round(dx_norm, PRECISION, out=dx_norm)
    return dx_norm


def _calc_std_mean(
    x: np.ndarray,
    f: np.ndarray,
    n_points: int = 10_000,
) -> float:
    from ..model import build_gp_matern

    np.random.seed(0)
    x_sample = np.random.uniform(x.min(0), x.max(0), size=(n_points, x.shape[1]))
    std_means = []
    for ls in np.logspace(-5, 5, 11):
        model = build_gp_matern(x=x, f=f, length_scale=ls, length_scale_bounds="fixed")
        std_means.append(model.predict(x_sample, return_std=True)[1].mean())
    return np.mean(std_means)


def _calc_std_max(
    x: np.ndarray,
    f: np.ndarray,
    n_restarts: int = 100,
) -> float:
    from scipy import optimize

    from ..model import build_gp_matern

    model = build_gp_matern(x=x, f=f, n_restarts_optimizer=5)

    def std(model, x):
        x = np.atleast_2d(x)
        _, std_pred = model.predict(x, return_std=True)
        return std_pred

    x_min = x.min(0)
    x_max = x.max(0)
    np.random.seed(0)
    x_inits = np.random.uniform(x_min, x_max, size=(n_restarts, x.shape[1]))
    best_x, best_f = None, np.inf
    for x_init in x_inits:
        res = optimize.minimize(
            lambda _: -std(model, _),
            x_init,
            bounds=list(zip(x_min, x_max)),
            method="L-BFGS-B",
        )
        if res.success and res.fun < best_f:
            best_f = res.fun
            best_x = res.x
    return best_x, -best_f


def plot(
    x: np.ndarray,
    triples: List[Tuple[int, int, int]],
    f: np.ndarray = None,
    title: str = "",
    size: Tuple[int, int] = (10, 6),
    dpi: int = 100,
    file_name: Optional[str] = None,
):
    import matplotlib.pyplot as plt

    plt.figure(figsize=size, dpi=dpi)

    [plt.plot(xt[:, 0], xt[:, 1]) for xt in x[triples]]
    plt.plot(x[:, 0], x[:, 1], ".", c="k")
    if f is not None:
        [
            plt.text(xi[0], xi[1], f"{i}|{fi[0]:.1f}", fontsize=8)
            for i, (xi, fi) in enumerate(zip(x, f))
        ]
    plt.xlabel("$x_1$")
    plt.ylabel("$x_2$")
    plt.title(title)
    plt.grid()
    if file_name:
        plt.savefig(file_name)
        plt.close()
    else:
        plt.show()


def collect_fixed_step(
    x: np.ndarray,
    max_step: float = 1.1,
    min_angle_deg: int = 90,
    symmetric: bool = False,
    normalize_distances: bool = True,
) -> List[Tuple[int, int, int]]:
    """
    - Iterative search algorithm
    - The number of triples is fixed and depends on parameters set (may be extreme in higher dimensions)
    - Generates triples with worse angle in average (less than 180)
    - Generates triples with better distances in average (points are closer to each other)
    - Nice for small samples at moderate dimensions
    """
    assert 0 <= max_step
    assert 0 <= min_angle_deg <= 180

    x = _normalize_x(x)
    # Calculate max min distance to determine step size
    distances = np.hypot.reduce(x[:, np.newaxis] - x, axis=2)
    if normalize_distances:
        distances = _normalize_distances(distances)
    np.round(distances, PRECISION, out=distances)
    # Remove duplicates
    distances[distances < np.finfo(float).eps] = np.inf
    max_min_distance = distances.min(0).max()
    step_size = max_min_distance * max_step

    # Find triples of points
    n = x.shape[0]
    triples: List[Tuple[int, int, int]] = []
    for i in np.arange(n):
        closest_idx = np.argwhere(distances[i] <= step_size).flatten()
        if closest_idx.size < 2:
            continue
        candidate_triples = []
        for i1, i3 in itertools.combinations(closest_idx, 2):
            candidate_triples.append((i1, i, i3))
        candidate_triples = np.array(candidate_triples)
        angles = _calc_angle(x, candidate_triples)
        triples.extend(candidate_triples[angles >= min_angle_deg])
    triples = np.array(triples)
    if symmetric:
        triples = np.vstack((triples, triples[:, [2, 1, 0]]))
    return triples


def collect_random(
    x: np.ndarray,
    n_triples: int,
    max_step: float = 1.0,
    step_multiplier: float = 1.2,
    min_angle_deg: int = 90,
    seed: int = 0,
    symmetric: bool = False,
):
    """
    - Iterative search algorithm
    - Can generate the specified number of triples
    - Generates triples with better angle in average (closer to 180)
    - Generates triples with worse distances in average (points are more distant)
    - Nice for huge samples at higher dimensions when the required number of fronts is known
    """
    assert 0 <= max_step
    np.random.seed(seed)

    x = _normalize_x(x)
    # Calculate max min distance to determine step size
    distances = np.hypot.reduce(x[:, np.newaxis] - x, axis=2)
    np.round(distances, PRECISION, out=distances)
    # Remove duplicates
    distances[distances < np.finfo(float).eps] = np.inf
    max_min_distance = distances.min(0).max()
    step_size = max_min_distance * max_step

    def _iterate_randomly(n_values):
        n = 0
        while True:
            has_neighbors_mask = np.sum(np.isfinite(distances), axis=1) > 1
            indices = np.arange(n_values)[has_neighbors_mask]
            np.random.shuffle(indices)
            if indices.size == 0:
                break
            for i in indices:
                n += 1  # should start from 1
                yield i, n

    # Set this parameter to 1 to select triples with best angles, but it may slow down
    # the algorithm and increase distances between the points of triples
    triples_per_point = int(np.clip(n_triples / x.shape[0], 1, np.inf)) + 1

    triples: Set[Tuple[int, int, int]] = set()
    random_idx = _iterate_randomly(x.shape[0])
    check_point = int(10 * max(x.shape[0], n_triples))
    while len(triples) < n_triples:
        # Select next point randomly and find neighbors
        try:
            i2, n = next(random_idx)
        except StopIteration:
            LOG.debug(
                f"WARN: {n_triples} triples requested but only {len(triples)} "
                f"collected, sample size {len(x)} is not enough"
            )
            break
        neighbors = np.argwhere(distances[i2] <= step_size).flatten()
        if neighbors.size > 1:
            # Form candidate triples starting from a random point
            candidate_triples = []
            for i1, i3 in itertools.combinations(neighbors, 2):
                candidate_triples.append((i1, i2, i3))
            candidate_triples = np.array(candidate_triples)
            # Choose fronts having maximum angle assumed by its points
            angles = _calc_angle(x, candidate_triples)
            sorted_idx = (-angles).argsort()
            for i in sorted_idx[:triples_per_point]:
                i1, i2, i3 = candidate_triples[i]
                # Do not select these 2 points next time e.g. in case if angle is less than required
                distances[i2, [i1, i3]] = np.inf
                if angles[i] < min_angle_deg:
                    break
                # Increase distances to reduce "weight" of that points next time
                distances[i1, i2] *= step_multiplier
                distances[i3, i2] *= step_multiplier
                triples.add((i1, i2, i3))
                if len(triples) == n_triples:
                    break
        # Increase step_size to avoid infinite loops
        if n % check_point == 0:
            step_size *= step_multiplier
    triples = np.array(list(triples))
    if symmetric:
        triples = np.vstack((triples, triples[:, [2, 1, 0]]))
    return triples


def collect_space_filling(
    x: np.ndarray,
    n_triples: int,
    min_angle_deg: int = 90,
    n_angle_ranges: int = 3,
    step_multiplier: float = 1.1,
    seed: int = 0,
    shuffle: bool = True,
    symmetric: bool = False,
):
    """
    Increasing step multiplier increases number of unique pairs but also increases the
    average distance between paired points.
    """
    assert (
        step_multiplier > 1
    ), "Step multiplier must be more than 1 to avoid infinite looping"
    np.random.seed(seed)

    x = _normalize_x(x)
    # Calculate max min distance to determine step size
    distances = np.hypot.reduce(x[:, np.newaxis] - x, axis=2)
    np.round(distances, PRECISION, out=distances)
    # Remove duplicates
    distances[distances < np.finfo(float).eps] = np.inf
    mean_distances = np.nanmean(np.nan_to_num(distances, posinf=np.nan), axis=1)

    def _iterate_randomly(n_values):
        indices = np.arange(n_values)
        if shuffle:
            np.random.shuffle(indices)
        while True:
            for i in indices:
                yield i

    angle_ranges = np.linspace(min_angle_deg, 180, n_angle_ranges + 1)
    angle_ranges = np.vstack([angle_ranges[:-1], angle_ranges[1:]]).T
    angle_ranges[-1, -1] += 1  # For strict inequality check

    n = x.shape[0]
    random_idx = _iterate_randomly(n)
    triples: List[Tuple[int, int, int]] = []

    while len(triples) < n_triples:
        # Select next point randomly and find the closest point
        i2 = next(random_idx)
        if not np.isfinite(distances[i2]).any():
            LOG.debug(
                f"WARN: {n_triples} triples requested but only {len(triples)} "
                f"collected, sample size {n} is not enough"
            )
            break
        i1 = np.argmin(distances[i2])
        distances[i2, i1] = np.inf
        # Build candidate triples from 2 closest points and all other points

        i3_list = np.delete(np.arange(n), [i2, i1])
        less_than_mean_distance = distances[i2, i3_list] <= mean_distances[i2]
        i3_list = i3_list[np.argwhere(less_than_mean_distance).flatten()]
        i2_list = np.tile(i2, len(i3_list))
        i1_list = np.tile(i1, len(i3_list))
        candidate_triples = np.vstack([i1_list, i2_list, i3_list]).T
        # Calculate angles to select the closest point within each angle range
        angles = _calc_angle(x, candidate_triples)
        for a_min, a_max in angle_ranges:
            best_idx = np.argwhere((a_min <= angles) * (angles < a_max)).flatten()
            # Point indices that satisfy the current angle range
            best_idx = candidate_triples[best_idx][:, -1]
            best_distances = distances[i2, best_idx]
            if not np.isfinite(best_distances).any():
                # Also checks for zero size of best_distances
                continue
            # Select the closest point in that angle range
            i3 = best_idx[best_distances.argmin()]
            triples.append((i1, i2, i3))
            # Increase distances so we do not choose the same points next time
            distances[i2, i3] *= step_multiplier
            if len(triples) == n_triples:
                break
    triples = np.array(triples)
    assert len(triples) == len(np.unique(triples, axis=0))
    if triples.size and symmetric:
        triples = np.vstack((triples, triples[:, [2, 1, 0]]))
    return triples


def extend(
    x: np.ndarray[float],
    f: np.ndarray[float],
    triples: np.ndarray[int],  # matrix of shape (n_triples, 3)
    n_splits_min: int = 1,
    n_splits_max: int = 3,
    n_splits_adaptive: bool = True,
    log_splits: bool = True,
    log_rate: float = 10.0,
    symmetric: bool = False,
) -> Tuple[np.ndarray[float], np.ndarray[float], np.ndarray[int]]:
    assert n_splits_max >= n_splits_min > 0

    # Pairs of step sizes for each triple - matrix of shape (n_triples, 2)
    dx_norm = _calc_dx_norm(x, triples)
    # Merge all triples into single 1D partially linear curve, such way we interpolate
    # all the new points in one call of np.interp
    zero_steps = np.zeros((len(dx_norm), 1))
    x_path_1D = np.cumsum(np.hstack([zero_steps, dx_norm]))
    f_path_1D = f[triples].flatten()

    # Estimate number of splits for each step depending on step size
    if n_splits_adaptive:
        min_step = dx_norm.min() / (n_splits_min + 1)
        max_step = dx_norm.max() / (n_splits_max + 1)
        mean_step = np.mean([min_step, max_step])
        n_ranges = np.floor(dx_norm / mean_step).astype(int)
        n_splits = np.clip(n_ranges - 1, n_splits_min, n_splits_max)
    else:
        n_splits_list = np.arange(n_splits_min, n_splits_max + 1)
        dx_ranges = np.linspace(dx_norm.min(), dx_norm.max(), len(n_splits_list) + 1)
        n_splits = np.empty_like(dx_norm, dtype=int)
        n_splits[:, 0] = np.argmax(dx_norm[:, [0]] <= dx_ranges, axis=1)
        n_splits[:, 1] = np.argmax(dx_norm[:, [1]] <= dx_ranges, axis=1)
        np.clip(n_splits - 1, 0, len(n_splits_list) - 1, out=n_splits)
        n_splits = n_splits_list[n_splits]

    weights: Dict[int, np.ndarray] = {}
    for i in np.unique(n_splits):
        w = np.linspace(0, 1, i + 2)[1:-1]
        if log_splits:
            w = 1.0 / (1.0 + np.exp(-log_rate * (w - 0.5)))
        weights[i] = np.vstack([w[::-1], w]).T

    triples_new = []
    x_new_1D = []
    x_new = []
    # Here we map ranges of triples to indices of new points generated within that
    # range. Multiple triples may share the same range, so we do not generate duplicated
    # points within such ranges and reuse the existing indices.
    new_idx_map = {}

    # Variable to increment new points indices
    i_new = x.shape[0]
    for i, ((t1, t2, t3), (n1, n2)) in enumerate(zip(triples, n_splits)):
        # Nodes of 1D partially linear curve corresponding to i-th triple
        x1_1D, x2_1D, x3_1D = x_path_1D[i * 3 : (i + 1) * 3]

        # Unique id of the 1st range from sorted indices of the first 2 points
        range1_id = (t1, t2) if t1 <= t2 else (t2, t1)
        # But first check if new points have not already been generated in that range.
        if range1_id not in new_idx_map:
            # Interpolate along the range in original x space to generate new points.
            # In linear case it's just np.linspace(x[t1], x[t2], 2 + n1)[1:-1]
            x_interp = weights[n1] @ x[[t1, t2]]
            x_new.extend(x_interp)

            # Interpolate along the range for 1D partially linear curve
            # In linear case it's just np.linspace(x1_1D, x2_1D, 2 + n1)[1:-1]
            x_interp_1D = weights[n1] @ [x1_1D, x2_1D]
            x_new_1D.extend(x_interp_1D)
            # Cache indices of newly generated points
            new_idx_map[range1_id] = list(range(i_new, i_new + n1))
            i_new += n1

        # Unique id of the 2nd range from sorted indices of the second 2 points
        range2_id = (t2, t3) if t2 <= t3 else (t3, t2)
        # But first check if new points have not already been generated in that range.
        if range2_id not in new_idx_map:
            # Interpolate along the range in original x space to generate new points.
            # In linear case it's just np.linspace(x[t2], x[t3], 2 + n2)[1:-1]
            x_interp = weights[n2] @ x[[t2, t3]]
            x_new.extend(x_interp)
            # Interpolate along the range for 1D partially linear curve
            # In linear case it's just np.linspace(x2_1D, x3_1D, 2 + n2)[1:-1]
            x_interp_1D = weights[n2] @ [x2_1D, x3_1D]
            x_new_1D.extend(x_interp_1D)
            # Cache indices of newly generated points
            new_idx_map[range2_id] = list(range(i_new, i_new + n2))
            i_new += n2

        path = [t1, *new_idx_map[range1_id], t2, *new_idx_map[range2_id], t3]
        # Add all possible triples along that path, triples may be duplicated
        # for some ranges but we will filter them out later since they are going to
        # have the same indices of newly generated points.
        triples_new.extend(np.vstack([path[:-2], path[1:-1], path[2:]]).T)

    f_new = np.interp(x_new_1D, x_path_1D, f_path_1D).reshape(-1, 1)

    triples_new = np.unique(triples_new, axis=0)
    if symmetric:
        triples_new = np.vstack((triples_new, triples_new[:, [2, 1, 0]]))

    x_ext = np.vstack([x, x_new])
    f_ext = np.vstack([f, f_new])
    triples_ext = np.vstack([triples, triples_new])
    return x_ext, f_ext, triples_ext



import datetime
import logging
from pathlib import Path

LOG_FOLDER = "logs"


def set_logger(filename: str = "", loglevel: int = logging.INFO) -> logging.Logger:
    if filename:
        filename += " "
    Path(LOG_FOLDER).mkdir(parents=True, exist_ok=True)
    path = Path(LOG_FOLDER) / f"{filename}{datetime.datetime.now():%Y.%m.%d %H.%M}.log"
    formatter = logging.Formatter(
        "[%(asctime)s] <%(levelname)s> (%(name)s) %(message)s"
    )

    handler_stdout = logging.StreamHandler()
    handler_stdout.setFormatter(formatter)
    handler_stdout.setLevel(loglevel)

    handler_file = logging.FileHandler(str(path), mode="w")
    handler_file.setFormatter(formatter)
    handler_file.setLevel(logging.DEBUG)

    root_logger = logging.getLogger()
    root_logger.addHandler(handler_file)
    root_logger.addHandler(handler_stdout)
    root_logger.setLevel(logging.DEBUG)
    return root_logger



import logging
import time
from typing import Callable, List, Tuple

import numpy as np
import pandas as pd
from bayes_opt import UtilityFunction
from bayes_opt.bayesian_optimization import BayesianOptimization
from bayes_opt.util import UtilityFunction, acq_max

from .constants import METRICS
from .model import build_gp_matern, models_estimate, models_select
from .problems.problem import ProblemBase
from .sample import DOE, SampleCached

LOG = logging.getLogger(__name__)


class Optimizer(BayesianOptimization):
    def __init__(
        self,
        problem: ProblemBase,
        init_xf: Tuple[np.ndarray, np.ndarray],
        metric: METRICS,
        predictor: Callable[[np.ndarray, np.ndarray], Tuple[float, float]],
        *args,
        **kwargs,
    ):
        self.problem = problem
        self.init_x, self.init_f = init_xf
        self.metric = metric
        self.predictor = predictor

        super().__init__(
            f=lambda **kwa: -problem.calc(self.space.params_to_array(kwa))[0][0],
            pbounds={
                f"x[{i}]": bound
                for i, bound in enumerate(zip(*problem.variable_bounds))
            },
            *args,
            **kwargs,
        )

        for x in self.init_x:
            # Do not register f values so that the
            # initial sample is printed to log
            self.probe(x)

        n_init = len(self.init_x)
        # Models history
        self.history_nu = [np.nan] * n_init
        self.history_length_scale = [np.nan] * n_init
        self.history_metric_values = [np.nan] * n_init
        # Values history
        self.history_x = [_.tolist() for _ in self.init_x]
        self.history_f = [_[0] for _ in self.init_f]
        self.history_f_model = [np.nan] * n_init
        self.history_time = [np.nan] * n_init

    def suggest(self, utility_function: UtilityFunction):
        time_start = time.time()

        x, neg_f = self._space.params, self._space.target.reshape(-1, 1)
        if self.metric == METRICS.NONE:
            best_model = build_gp_matern(x, neg_f, nu=2.5, n_restarts_optimizer=5)
            best_metric_value = np.nan
        elif self.metric == METRICS.PREDICTED:
            # We used normal f values for sample feature estimation
            nu = self.predictor(x, -neg_f)
            best_model = build_gp_matern(x, neg_f, nu=nu, n_restarts_optimizer=5)
            best_metric_value = np.nan
        else:
            models, metrics = models_estimate(self.problem, self.metric, x, neg_f)
            best_index = models_select(metrics, self.metric)
            best_model, best_metric_value = models[best_index], metrics[best_index]
        # Models history
        self.history_nu.append(best_model.kernel_.nu)
        self.history_length_scale.append(best_model.kernel_.length_scale)
        self.history_metric_values.append(best_metric_value)

        # Finding argmax of the acquisition function.
        suggestion_x: np.ndarray = acq_max(
            ac=utility_function.utility,
            gp=best_model,
            constraint=self.constraint,
            y_max=self._space.target.max(),
            bounds=self._space.bounds,
            random_state=self._random_state,
        )
        suggestion_f = float(self.problem.calc(suggestion_x))
        suggestion_f_model = float(-best_model.predict([suggestion_x]))

        # Values history
        self.history_x.append(suggestion_x.tolist())
        self.history_f.append(suggestion_f)
        self.history_f_model.append(suggestion_f_model)
        self.history_time.append(time.time() - time_start)

        return self._space.array_to_params(suggestion_x)


def solve_problem(
    problem: ProblemBase,
    n_init: int,
    n_iter: int,
    n_seeds: int,
    metric: METRICS,
    predictor: Callable[[np.ndarray, np.ndarray], Tuple[float, float]] = None,
) -> pd.DataFrame:
    solution_f = problem.solutions[0].f[0] if len(problem.solutions) else np.nan
    n_total = n_init + n_iter

    data_frames: List[pd.DataFrame] = []
    for i in range(n_seeds):
        init_sample = SampleCached(
            problem,
            size=n_init,
            doe=DOE.LHS,
            options={"seed": i},
        )
        optimizer = Optimizer(
            problem=problem,
            init_xf=init_sample.xf,
            metric=metric,
            predictor=predictor,
            # Parent args:
            random_state=i,
            allow_duplicate_points=True,
        )
        optimizer.maximize(init_points=0, n_iter=n_iter)
        best_idx = np.argmin(optimizer.history_f)
        x_best = optimizer.history_x[best_idx]
        f_best = optimizer.history_f[best_idx]
        data_frames.append(
            pd.DataFrame(
                {
                    # Problem definition
                    "problem": [problem.NAME] * n_total,
                    "dim": [problem.dim_x] * n_total,
                    "n_init": [n_init] * n_total,
                    "n_iter": [n_iter] * n_total,
                    "metric": [metric] * n_total,
                    "solution_f": [solution_f] * n_total,
                    # Current run info
                    "seed": [i] * n_total,
                    "iteration": list(range(1, n_init + n_iter + 1)),
                    # Models history
                    "nu": optimizer.history_nu,
                    "length_scale": optimizer.history_length_scale,
                    "metric_values": optimizer.history_metric_values,
                    # Values history
                    "x": optimizer.history_x,
                    "f": optimizer.history_f,
                    "f_model": optimizer.history_f_model,
                    "time": optimizer.history_time,
                    "x_best": [x_best] * n_total,
                    "f_best": [f_best] * n_total,
                }
            )
        )
        LOG.info(f"{problem.id} [{i+1}/{n_seeds}]\tx*={x_best}; " f"f(x*)={f_best}")
    return pd.concat(data_frames, ignore_index=True)



from ..functions import levy5
from ..problem import ProblemBase


class Problem(ProblemBase):
    def init(self):
        self.add_variable(bounds=(-2, 2), initial_guess=1)
        self.add_variable(bounds=(-2, 2), initial_guess=1)
        self.add_objective()
        self.add_solution(x=[-1.3086, -1.4248], f=-176.1376)

    def calc_objectives(self, x):
        return [levy5(*x)]



from ..functions import powell
from ..problem import ProblemBase


class Problem(ProblemBase):
    def __init__(self, dim_x=4):
        self.__dim_x = dim_x
        if self.__dim_x % 4 != 0:
            raise ValueError("The problem dim_x should be multiple of 4")

    def init(self):
        for _ in range(self.__dim_x):
            self.add_variable(bounds=(-4, 5), initial_guess=4)
        self.add_objective()
        self.add_solution(x=[0.0] * self.__dim_x, f=0)

    def calc_objectives(self, x):
        return [powell(x)]



from pathlib import Path

from ._problem import ProblemBBOB

_all = []
__all__ = [
    name.stem
    for name in Path(__file__).parent.iterdir()
    if name.suffix == ".py" and not name.stem.startswith("_")
]

for _name in __all__:
    try:
        exec(f"from . import {_name}")
        exec(f"_all.append({_name})")
    except Exception as ex:
        print(f"Failed to import {_name}: {ex}")

# Sort by name
_all = sorted(_all, key=lambda problem: problem.Problem().NAME)



import copy
from collections.abc import Sequence
from dataclasses import dataclass
from typing import List, Optional, Tuple

import numpy as np

from ..constants import PROBLEM_FEATURE


def check_bounds(bounds: Sequence, name: str) -> Tuple[float, float]:
    if not isinstance(bounds, Sequence) or len(bounds) != 2:
        raise ValueError(f"Wrong bounds structure of {name}")
    bounds_float = np.array(bounds, dtype=float)
    if not np.isfinite(bounds_float[0]):
        bounds_float[0] = -np.inf
    if not np.isfinite(bounds_float[1]):
        bounds_float[1] = np.inf
    if bounds_float[0] > bounds_float[1]:
        raise ValueError(f"Invalid bound values {bounds_float.tolist()} of {name}")
    return tuple(bounds_float.tolist())


@dataclass
class Solution:
    x: np.ndarray[float]
    f: Optional[np.ndarray[float]] = None
    c: Optional[np.ndarray[float]] = None

    def __post_init__(self):
        self.x = np.atleast_1d(self.x).astype(float)
        if self.f is not None:
            self.f = np.atleast_1d(self.f).astype(float)
        if self.c is not None:
            self.c = np.atleast_1d(self.c).astype(float)

    def __str__(self):
        return (
            f"x: {np.array_str(self.x)}; "
            f"f: {np.array_str(self.f)}; "
            f"c: {np.array_str(self.c)}; "
        )


@dataclass
class Variable:
    name: str
    bounds: Tuple[float, float] = (-np.inf, np.inf)
    initial_guess: Optional[float] = None

    def __post_init__(self):
        self.bounds = check_bounds(self.bounds, str(self))
        if self.initial_guess is not None:
            self.initial_guess = float(self.initial_guess)
            if not (self.bounds[0] <= self.initial_guess <= self.bounds[1]):
                raise ValueError(f"Invalid initial guess of {self}")

    def set_initial_guess(self, initial_guess):
        if initial_guess is not None:
            self.initial_guess = float(initial_guess)
        else:
            self.initial_guess = None

    def set_bounds(self, bounds):
        self.bounds = check_bounds(bounds, str(self))

    def __str__(self) -> str:
        return f"Variable `{self.name}` in {self.bounds} {self.initial_guess}"


@dataclass
class Objective:
    name: str

    def __str__(self) -> str:
        return f"Objective `{self.name}`"


@dataclass
class Constraint:
    name: str
    bounds: Tuple[float, float] = (-np.inf, np.inf)

    def __post_init__(self):
        self.bounds = check_bounds(self.bounds, str(self))

    def set_bounds(self, bounds):
        self.bounds = check_bounds(bounds, str(self))

    def __str__(self) -> str:
        return f"Constraint `{self.name}` in {self.bounds}"


class ProblemMeta(type):
    def __call__(cls, *args, **kwargs):
        problem: ProblemBase = type.__call__(cls, *args, **kwargs)  # overwritten
        problem._initialize()  # fixed method
        problem.init()  # overwritten
        # Number of variables is available after prepare problem
        problem.step = [0] * problem.dim_x
        problem.shift = [0] * problem.dim_x
        problem.rotation["matrix"] = [[np.nan] * problem.dim_x] * problem.dim_x
        return problem


class ProblemBase(metaclass=ProblemMeta):
    def _initialize(self):
        self.NAME: str = self.__module__.replace("boela.problems.", "")

        self.calls_count = 0
        self.variables: List[Variable] = []
        self.objectives: List[Objective] = []
        self.constraints: List[Constraint] = []
        self.solutions: List[Solution] = []
        self._history_cache: List[List[float]] = []

        self.scale = (0, 1)
        self.noise = 0
        self.step = []
        self.shift = []
        self.rotation = {"angle": 0, "matrix": [[]]}

    @property
    def id(self):
        return f"{self.NAME}_{self.dim_x}"

    # USER SIDE METHODS

    def init(self):
        raise Exception("Problem was not initialized")

    def calc_objectives(self, x):
        raise Exception("Objectives was not defined")

    def calc_constraints(self, x):
        raise Exception("Constraints was not defined")

    # EVALUATION METHODS

    def _calc_f(self, x):
        return np.array([self.calc_objectives(xi) for xi in x])

    def _calc_c(self, x):
        return np.array([self.calc_constraints(xi) for xi in x])

    def print_iters(self, step=1):
        self.i_print = lambda n, i: [
            print(_, end="'") for _ in range(n + 1, n + i + 1) if _ % step == 0
        ]
        return self

    def calc(self, x):
        x = np.atleast_2d(x)
        if x.shape[1] != self.dim_x:
            raise Exception("Wrong dimension of input points: " + str(x))

        batch_size = x.shape[0]
        designs = np.zeros((batch_size, self.dim_xfc))
        designs[:, : self.dim_x] = x
        if self.dim_f > 0:
            designs[:, self.dim_x : self.dim_x + self.dim_f] = self._calc_f(x)
        if self.dim_c > 0:
            designs[:, self.dim_x + self.dim_f :] = self._calc_c(x)

        getattr(self, "i_print", lambda *_: _)(self.calls_count, batch_size)

        self.calls_count += batch_size
        self._history_cache.extend(designs.tolist())
        return designs[:, self.dim_x :]

    # ADD METHODS

    def add_variable(self, bounds, initial_guess=None, name=None):
        assert not self.solutions, "Solutions can not be added before the variables"
        if not name:
            name = f"x[{self.dim_x + 1}]"
        self.variables.append(Variable(name, bounds, initial_guess))

    def add_objective(self, name=None):
        assert not self.solutions, "Solutions can not be added before the objectives"
        if not name:
            name = f"f[{self.dim_f + 1}]"
        self.objectives.append(Objective(name))

    def add_constraint(self, bounds, name=None):
        assert not self.solutions, "Solutions can not be added before the constraints"
        if not name:
            name = f"c[{self.dim_c + 1}]"
        self.constraints.append(Constraint(name, bounds))

    def add_solution(
        self,
        x: Sequence,
        f: Optional[Sequence] = None,
        c: Optional[Sequence] = None,
    ):
        x = np.atleast_1d(x)
        assert len(x) == self.dim_x
        if f is not None:
            f = np.atleast_1d(f)
            assert len(f) == self.dim_f
        if c is not None:
            c = np.atleast_1d(c)
            assert len(c) == self.dim_c
        self.solutions.append(Solution(x=x, f=f, c=c))
        if self.dim_f == 1 and len(self.solutions) > 1:
            f_values = [_.f[0] for _ in self.solutions]
            assert np.all(f_values[0] == f_values)

    # SIZE METHODS

    @property
    def dim_x(self):
        return len(self.variables)

    @property
    def dim_f(self):
        return len(self.objectives)

    @property
    def dim_c(self):
        return len(self.constraints)

    @property
    def dim_fc(self):
        return self.dim_f + self.dim_c

    @property
    def dim_xfc(self):
        return self.dim_x + self.dim_f + self.dim_c

    # NAMES METHODS

    @property
    def variable_names(self):
        return [_.name for _ in self.variables]

    @property
    def constraint_names(self):
        return [_.name for _ in self.constraints]

    @property
    def objective_names(self):
        return [_.name for _ in self.objectives]

    # BOUNDS AND INITIAL GUESS METHODS

    @property
    def variable_bounds(self):
        return list(zip(*[_.bounds for _ in self.variables]))

    @property
    def constraint_bounds(self):
        return list(zip(*[_.bounds for _ in self.constraints]))

    @property
    def initial_guess(self):
        return [_.initial_guess for _ in self.variables]

    # HISTORY METHODS

    @property
    def history(self):
        return np.array(self._history_cache)

    def clear_history(self):
        self._history_cache = []
        self.calls_count = 0

    def modify(
        self,
        noise: float = 0,
        step=0,
        scale=(0, 1),
        shift=0,
        rotation=0,
        seed=None,
    ):
        problem = copy.deepcopy(self)
        problem.clear_history()

        # Modify objective function
        define_objectives = problem.calc_objectives

        # Add noise to objective function
        if noise != 0:
            assert noise > 0
            problem.noise = noise

            def noise_wrapper(function):
                def wrapped(x):
                    result = np.array(function(x))
                    return result * np.random.uniform(1, 1 + noise)

                return wrapped

            define_objectives = noise_wrapper(define_objectives)

        # Shift objective function
        if shift != 0:
            assert shift > 0
            if np.ndim(shift) == 0:
                if seed is not None:
                    np.random.seed(seed)
                # x_domain = np.diff(problem.variables_bounds, axis=0).ravel()
                # shift = np.random.uniform(low=-shift * x_domain, high=shift * x_domain, size=problem.size_x)
                direction = np.random.choice((-1, 1), problem.dim_x)
                max_steps = np.diff(problem.variable_bounds, axis=0)[0]
                shift = shift * max_steps * direction
            elif np.ndim(shift) > 1:
                shift = np.array(shift).flatten()
            assert len(shift) == self.dim_x, f"Wrong shape of shift"

            problem.shift = shift

            def shift_wrapper(function):
                return lambda x: function(x + shift)

            define_objectives = shift_wrapper(define_objectives)

        # Rotate objective function
        if rotation != 0:
            if isinstance(rotation, dict):
                if "angle" not in rotation or "matrix" not in rotation:
                    raise AttributeError("Wrong value of rotation.")
                if np.array(rotation["matrix"]).shape != (
                    problem.dim_x,
                    problem.dim_x,
                ):
                    raise AttributeError("Wrong value of rotation matrix.")
                problem.rotation = rotation
            elif np.array(rotation).ndim == 0:
                if seed is not None:
                    np.random.seed(seed)
                # Vectors of the rotation plane
                v1 = np.random.rand(problem.dim_x)
                v2 = np.random.rand(problem.dim_x)
                # Gram-Schmidt orthogonalization
                n1 = v1 / np.linalg.norm(v1)
                v2 = v2 - np.dot(n1, v2) * n1
                n2 = v2 / np.linalg.norm(v2)
                # Rotation matrix preparation
                # Normalize angle since the area out of domain depends on dimensionality
                angle = rotation / problem.dim_x
                identity = np.identity(problem.dim_x)
                generator2 = (np.outer(n2, n1) - np.outer(n1, n2)) * np.sin(angle)
                generator3 = (np.outer(n1, n1) + np.outer(n2, n2)) * (np.cos(angle) - 1)
                rot_matrix = identity + generator2 + generator3
                problem.rotation = {"angle": angle, "matrix": rot_matrix.tolist()}

            x_center = np.mean(problem.variable_bounds, axis=0)

            def rotation_wrapper(function):
                return lambda x: function(
                    np.dot(x - x_center / 2.0, problem.rotation["matrix"])
                    + x_center / 2.0
                )

            define_objectives = rotation_wrapper(define_objectives)

        # Scale objective function to mean 0 and std 1 (if current scale differs)
        if scale != (0, 1):
            problem.scale = scale

            def scale_wrapper(function):
                return lambda x: np.array(function(x) - scale[0]) / scale[1]

            define_objectives = scale_wrapper(define_objectives)

        # Set step for problem inputs
        if isinstance(step, (list, tuple, np.ndarray)):
            assert (
                len(step) == problem.dim_x
            ), "Resolution should be an array of length size_x or double"
        else:
            step = [step] * problem.dim_x
        if np.any(np.array(step) > 0):
            problem.step = step

            def step_wrapper(function):
                def stepped_function(x):
                    for i, res in enumerate(problem.step):
                        x[i] = res * np.round(x[i] / res) if res > 0 else x[i]
                    return function(x)

                return stepped_function

            define_objectives = step_wrapper(define_objectives)

        problem.calc_objectives = define_objectives
        return problem

    def sample_ff(self, size, seed=None, random=True):
        from . import doe

        if seed is not None:
            np.random.seed(seed)
        lb, ub = np.array(self.variable_bounds)
        x = lb + doe.full_factorial(self.dim_x, size, random) * (ub - lb)
        return x, self.calc(x)

    def sample_grid(self, size, seed=None, random_shift=0, partial=False):
        from . import doe

        if seed is not None:
            np.random.seed(seed)
        lb, ub = np.array(self.variable_bounds)
        x = lb + doe.cells_grid(self.dim_x, size, random_shift) * (ub - lb)
        if partial and size is not None and size < len(x):
            x = x[np.random.choice(np.arange(len(x)), size, replace=False)]
        return x, self.calc(x)

    def sample_rand(self, size, seed=None):
        if seed is not None:
            np.random.seed(seed)
        lb, ub = np.array(self.variable_bounds)
        x = lb + np.random.rand(size, self.dim_x) * (ub - lb)
        return x, self.calc(x)

    def sample_lhs(self, size, seed=None, centered=False):
        from scipy.stats import qmc

        sampler = qmc.LatinHypercube(d=self.dim_x, seed=seed, centered=centered)
        x = qmc.scale(sampler.random(size), *self.variable_bounds)
        return x, self.calc(x)

    def features(self, feature: PROBLEM_FEATURE, *args, x=None, f=None, **kwargs):
        from ..ela import analyze_problem

        if feature not in PROBLEM_FEATURE:
            raise ValueError(f"Features of type {feature} require sample")
        sample_required = PROBLEM_FEATURE.sample_required(feature)
        bounds_required = PROBLEM_FEATURE.bounds_required(feature)
        if sample_required and (x is None or f is None):
            raise ValueError(f"Sample is required for `{feature}` features")
        lb, ub = self.variable_bounds
        # Flacco for list bound and fail with tuples
        lb, ub = list(lb), list(ub)
        calc = self.calc_objectives
        if sample_required and bounds_required:
            return analyze_problem(
                feature, x, f, calc, self.dim_x, lb, ub, *args, **kwargs
            )
        elif sample_required and not bounds_required:
            return analyze_problem(feature, x, f, calc, *args, **kwargs)
        elif not sample_required and bounds_required:
            return analyze_problem(feature, calc, self.dim_x, lb, ub, *args, **kwargs)
        else:
            return analyze_problem(feature, calc, self.dim_x, *args, **kwargs)

    def plot(
        self,
        n_nodes_x1=50,
        n_nodes_x2=50,
        n_nodes_f=50,
        contour=False,
        wireframe=False,
        sample_x: np.ndarray = None,
        sample_f: np.ndarray = None,
        size=(6, 5),
        file_name=None,
        dpi=100,
    ):
        from ..model import plot_surface

        if self.dim_x != 2:
            raise Exception("For 2d problems only!")
        plot_surface(
            func=self.calc,
            lower_bound=self.variable_bounds[0],
            upper_bound=self.variable_bounds[1],
            n_nodes_x1=n_nodes_x1,
            n_nodes_x2=n_nodes_x2,
            n_nodes_f=n_nodes_f,
            contour=contour,
            sample_x=sample_x,
            sample_f=sample_f,
            wireframe=wireframe,
            file_name=file_name,
            title=self.NAME,
            size=size,
            dpi=dpi,
        )



from ..functions import hosaki
from ..problem import ProblemBase


class Problem(ProblemBase):
    def init(self):
        self.add_variable(bounds=(0, 10), initial_guess=10)
        self.add_variable(bounds=(0, 10), initial_guess=10)
        self.add_objective()
        self.add_solution(x=[4, 2], f=-2.34589)

    def calc_objectives(self, x):
        return [hosaki(*x)]



from ..functions import sargan
from ..problem import ProblemBase


class Problem(ProblemBase):
    def __init__(self, dim_x=2):
        self.__dim_x = dim_x

    def init(self):
        for _ in range(self.__dim_x):
            self.add_variable(bounds=(-100, 100), initial_guess=100)
        self.add_objective()
        self.add_solution(x=[0.0] * self.__dim_x, f=0)

    def calc_objectives(self, x):
        return [sargan(x)]



import warnings

import numpy as np
from pflacco import classical_ela_features, local_optima_network_features, misc_features


# Classical ELA Features
def ela_meta(x, f):
    return classical_ela_features.calculate_ela_meta(X=x, y=f)


def ela_distr(x, f, *args, **kwargs):
    res = classical_ela_features.calculate_ela_distribution(x, f, *args, **kwargs)
    # Otherwise it fails with error on saving to json :
    # Object of type int32 is not JSON serializable
    res["ela_distr.number_of_peaks"] = float(res["ela_distr.number_of_peaks"])
    return res


def ela_level(x, f, *args, **kwargs):
    return classical_ela_features.calculate_ela_level(x, f, *args, **kwargs)


# Cell Mapping Features
def cm_angle(x, f, lb, ub, *args, **kwargs):
    return classical_ela_features.calculate_cm_angle(x, f, lb, ub, *args, **kwargs)


def cm_conv(x, f, lb, ub, *args, **kwargs):
    return classical_ela_features.calculate_cm_conv(x, f, lb, ub, *args, **kwargs)


def cm_grad(x, f, lb, ub, *args, **kwargs):
    return classical_ela_features.calculate_cm_grad(x, f, lb, ub, *args, **kwargs)


def cm_limo(x, f, lb, ub, *args, **kwargs):
    return classical_ela_features.calculate_limo(x, f, lb, ub, *args, **kwargs)


# Dispersion Features
def disp(x, f, *args, **kwargs):
    # It spams warnings for small samples:
    # - RuntimeWarning: Mean of empty slice
    # - RuntimeWarning: invalid value encountered in scalar divide
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        return classical_ela_features.calculate_dispersion(x, f, *args, **kwargs)


# Information Content-Based Features
def ic(x, f, *args, **kwargs):
    try:
        return classical_ela_features.calculate_information_content(
            x, f, seed=0, *args, **kwargs
        )
    except Exception as ex:
        warnings.warn(f"Failed to calculate IC features {ex}")
        return {
            "ic.h_max": np.nan,
            "ic.eps_s": np.nan,
            "ic.eps_max": np.nan,
            "ic.eps_ratio": np.nan,
            "ic.m0": np.nan,
            "ic.costs_runtime": np.nan,
        }


# Nearest Better Features
def nbc(x, f, *args, **kwargs):
    return classical_ela_features.calculate_nbc(x, f, *args, **kwargs)


# Principal Components Features
def pca(x, f, *args, **kwargs):
    return classical_ela_features.calculate_pca(x, f, *args, **kwargs)


# Fitness Distance Correlation features
def fdc(x, f, *args, **kwargs):
    return misc_features.calculate_fitness_distance_correlation(x, f, *args, **kwargs)


# Feature sets that do require additional function evaluations
def ela_local(x, f, calc, dim, lb, ub, *args, **kwargs):
    return classical_ela_features.calculate_ela_local(
        x, f, calc, dim, lb, ub, seed=0, *args, **kwargs
    )


def ela_curv(x, f, calc, dim, lb, ub, *args, **kwargs):
    return classical_ela_features.calculate_ela_curvate(
        x, f, calc, dim, lb, ub, *args, **kwargs
    )


def ela_conv(x, f, calc, *args, **kwargs):
    return classical_ela_features.calculate_ela_conv(
        x, f, calc, seed=0, *args, **kwargs
    )


def lon(calc, dim, lb, ub, *args, **kwargs):
    nodes, edges = local_optima_network_features.compute_local_optima_network(
        calc, dim, lb, ub, seed=0, *args, **kwargs
    )
    return local_optima_network_features.calculate_lon_features(nodes, edges)


def hill(calc, dim, lb, ub, *args, **kwargs):
    return misc_features.calculate_hill_climbing_features(
        calc, dim, lb, ub, seed=0, *args, **kwargs
    )


def grad(calc, dim, lb, ub, *args, **kwargs):
    return misc_features.calculate_gradient_features(
        calc, dim, lb, ub, seed=0, *args, **kwargs
    )


def scale(calc, dim, lb, ub, *args, **kwargs):
    return misc_features.calculate_length_scales_features(
        calc, dim, lb, ub, seed=0, *args, **kwargs
    )


def sobol(calc, dim, lb, ub, *args, **kwargs):
    return misc_features.calculate_sobol_indices_features(
        calc, dim, lb, ub, seed=0, *args, **kwargs
    )
